from easyprocess import EasyProcess

from pyscreenshot.plugins.backend import UNKNOWN_VERSION, CBackend
from pyscreenshot.tempexport import read_prog_img
from pyscreenshot.util import platform_is_osx

PROGRAM = "screencapture"
# https://ss64.com/osx/screencapture.html
#  By default screenshots are saved as .png files,


class ScreencaptureError(Exception):
    pass


class ScreencaptureWrapper(CBackend):
    name = "mac_screencapture"
    is_subprocess = True

    def grab(self, bbox=None):
        if not platform_is_osx():
            raise ScreencaptureError("This backend runs only on Darwin")

        command = [PROGRAM, "-x"]
        if bbox:
            width = bbox[2] - bbox[0]
            height = bbox[3] - bbox[1]
            command += ["-R{},{},{},{}".format(bbox[0], bbox[1], width, height)]
        im = read_prog_img(command)
        return im

    def backend_version(self):
        p = EasyProcess([PROGRAM, "-help"])
        p.enable_stdout_log = False
        p.enable_stderr_log = False
        p.call()
        if p.return_code == 0:
            return UNKNOWN_VERSION



import logging
import os

from easyprocess import EasyProcess
from PIL import Image

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.tempexport import RunProgError

log = logging.getLogger(__name__)


PROGRAM = "ksnip"


class KsnipWrapper(CBackend):
    name = "ksnip"
    is_subprocess = True

    def grab(self, bbox=None):
        cmd = [PROGRAM, "--fullscreen", "--save"]
        p = EasyProcess(cmd)
        p.call()
        if p.return_code != 0:
            raise RunProgError(p.stderr)
        lastline = p.stdout.splitlines()[-1]
        if "Image Saved" not in lastline:
            raise RunProgError(p.stderr)
        filename = lastline.split()[-1]
        im = Image.open(filename)
        os.remove(filename)

        # TODO: bbox param
        if bbox:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        for line in EasyProcess([PROGRAM, "--version"]).call().stderr.splitlines():
            if "version" in line.lower():
                return line.split()[-1]



import cv2.typing
import typing


# Enumerations
TraitAs_TENSOR: int
TRAIT_AS_TENSOR: int
TraitAs_IMAGE: int
TRAIT_AS_IMAGE: int
TraitAs = int
"""One of [TraitAs_TENSOR, TRAIT_AS_TENSOR, TraitAs_IMAGE, TRAIT_AS_IMAGE]"""



# Classes
class PyParams:
    # Functions
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, tag: str, model_path: str) -> None: ...

    def cfgMeanStd(self, layer_name: str, m: cv2.typing.Scalar, s: cv2.typing.Scalar) -> PyParams: ...

    def cfgNormalize(self, layer_name: str, flag: bool) -> PyParams: ...



# Functions
def params(tag: str, model_path: str) -> PyParams: ...





import logging

from pyscreenshot.plugins.backend import UNKNOWN_VERSION, CBackend
from pyscreenshot.tempexport import read_func_img

log = logging.getLogger(__name__)


class GnomeDBusError(Exception):
    pass


class GnomeDBusWrapper(CBackend):
    name = "gnome_dbus"
    is_subprocess = True

    def grab(self, bbox=None):
        im = read_func_img(self._grab_to_file, bbox)
        return im

    def _grab_to_file(self, filename, bbox=None):
        has_jeepney = False
        try:
            # from jeepney import new_method_call
            from jeepney.io.blocking import open_dbus_connection  # type: ignore
            from jeepney.wrappers import MessageGenerator  # type: ignore
            from jeepney.wrappers import new_method_call

            has_jeepney = True
        except ImportError:
            pass

        if not has_jeepney:
            raise GnomeDBusError("jeepney library is missing")

        class Screenshot(MessageGenerator):
            interface = "org.gnome.Shell.Screenshot"

            def __init__(
                self,
                object_path="/org/gnome/Shell/Screenshot",
                bus_name="org.gnome.Shell.Screenshot",
            ):
                super().__init__(object_path=object_path, bus_name=bus_name)

            def Screenshot(self, include_cursor, flash, filename):
                return new_method_call(
                    self, "Screenshot", "bbs", (include_cursor, flash, filename)
                )

            def ScreenshotArea(self, x, y, width, height, flash, filename):
                return new_method_call(
                    self,
                    "ScreenshotArea",
                    "iiiibs",
                    (x, y, width, height, flash, filename),
                )

        # https://jeepney.readthedocs.io/en/latest/integrate.html
        connection = open_dbus_connection(bus="SESSION")
        dbscr = Screenshot()
        if bbox:
            msg = dbscr.ScreenshotArea(
                bbox[0],
                bbox[1],
                bbox[2] - bbox[0],
                bbox[3] - bbox[1],
                False,
                filename,
            )
        else:
            msg = dbscr.Screenshot(False, False, filename)
        reply = connection.send_and_get_reply(msg)
        result = reply.body[0]
        if not result:
            raise GnomeDBusError("dbus error: %s %s" % (msg, result))

    def backend_version(self):
        return UNKNOWN_VERSION



import cv2
import cv2.typing
import typing


# Classes
class BarcodeDetector(cv2.GraphicalCodeDetector):
    # Functions
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, prototxt_path: str, model_path: str) -> None: ...

    @typing.overload
    def decodeWithType(self, img: cv2.typing.MatLike, points: cv2.typing.MatLike) -> tuple[bool, typing.Sequence[str], typing.Sequence[str]]: ...
    @typing.overload
    def decodeWithType(self, img: cv2.UMat, points: cv2.UMat) -> tuple[bool, typing.Sequence[str], typing.Sequence[str]]: ...

    @typing.overload
    def detectAndDecodeWithType(self, img: cv2.typing.MatLike, points: cv2.typing.MatLike | None = ...) -> tuple[bool, typing.Sequence[str], typing.Sequence[str], cv2.typing.MatLike]: ...
    @typing.overload
    def detectAndDecodeWithType(self, img: cv2.UMat, points: cv2.UMat | None = ...) -> tuple[bool, typing.Sequence[str], typing.Sequence[str], cv2.UMat]: ...






import os.path
from tempfile import TemporaryDirectory

from easyprocess import EasyProcess
from PIL import Image


class RunProgError(Exception):
    pass


def read_func_img(func, bbox=None):
    with TemporaryDirectory(prefix="pyscreenshot") as tmpdirname:
        filename = os.path.join(tmpdirname, "screenshot.png")
        func(filename, bbox)
        im = Image.open(filename)
        return im


def read_prog_img(cmd):
    def run_prog(filename, bbox=None):
        p = EasyProcess(cmd + [filename])
        p.call()
        if p.return_code != 0:
            raise RunProgError(p.stderr)

    im = read_func_img(run_prog)
    return im



import logging

from entrypoint2 import entrypoint

from pyscreenshot.loader import backend_version2

log = logging.getLogger(__name__)


@entrypoint
def main(backend):
    """Print pyscreenshot back-end version.

    :param backend: back-end (example:scrot, wx,..)
    """
    backend = backend if backend else None

    try:
        v = backend_version2(backend)
    except Exception as e:
        log.warning(e)
        v = ""
    print(v)



# Xlib.__init__ -- glue for Xlib package
#
#    Copyright (C) 2000-2002 Peter Liljenberg <petli@ctrl-c.liu.se>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__version__ = (0, 15)

__version_extra__ = ''

__version_string__ = '.'.join(map(str, __version__)) + __version_extra__

__all__ = [
    'X',
    'XK',
    'Xatom',
    'Xcursorfont',
    'Xutil',
    'display',
    'error',
    'rdb',
    # Explicitly exclude threaded, so that it isn't imported by
    #  from Xlib import *
    ]



"Create screenshot of xmessage with Xvfb"
from time import sleep

from easyprocess import EasyProcess
from pyvirtualdisplay import Display

import pyscreenshot as ImageGrab

with Display(size=(100, 60)) as disp:  # start Xvfb display
    # display is available
    with EasyProcess(["xmessage", "hello"]):  # start xmessage
        sleep(1)  # wait for displaying window
        img = ImageGrab.grab()
img.save("xmessage.png")



import io
import logging

from PIL import Image

from pyscreenshot.plugins.backend import CBackend

log = logging.getLogger(__name__)


class PySide2BugError(Exception):
    pass


app = None


class PySide2GrabWindow(CBackend):
    name = "pyside2"

    def grab_to_buffer(self, buff, file_type="png"):
        from PySide2 import QtCore, QtGui, QtWidgets  # type: ignore

        QApplication = QtWidgets.QApplication
        QBuffer = QtCore.QBuffer
        QIODevice = QtCore.QIODevice
        QScreen = QtGui.QScreen
        # QPixmap = self.PySide2.QtGui.QPixmap

        global app
        if not app:
            app = QApplication([])
        qbuffer = QBuffer()
        qbuffer.open(QIODevice.ReadWrite)
        QScreen.grabWindow(
            QApplication.primaryScreen(), QApplication.desktop().winId()
        ).save(qbuffer, file_type)
        # https://stackoverflow.com/questions/52291585/pyside2-typeerror-bytes-object-cannot-be-interpreted-as-an-integer
        buff.write(qbuffer.data().data())
        qbuffer.close()

    def grab(self, bbox=None):
        strio = io.BytesIO()
        self.grab_to_buffer(strio)
        strio.seek(0)
        im = Image.open(strio)
        if bbox:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        import PySide2

        return PySide2.__version__



# import atexit

from PIL import Image

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.util import py_minor

# https://python-mss.readthedocs.io/examples.html


class MssError(Exception):
    pass


sct = None

# not working without xrandr extension
# only bits_per_pixel == 32 is supported


class MssWrapper(CBackend):
    name = "mss"

    def grab(self, bbox=None):
        if py_minor() < 5:
            raise MssError()
        import mss  # type: ignore

        # atexit.register(sct.close())

        global sct
        if not sct:
            sct = mss.mss()

        # with self.mss.mss() as sct:
        if bbox:
            monitor = bbox
        else:
            monitor = sct.monitors[0]
        sct_img = sct.grab(monitor)

        im = Image.frombytes("RGB", sct_img.size, sct_img.bgra, "raw", "BGRX")
        # The same, but less efficient:
        # im = Image.frombytes('RGB', sct_img.size, sct_img.rgb)
        return im

    def backend_version(self):
        import mss

        return mss.__version__



# PyScreeze - PyScreeze is a simple, cross-platform screenshot module for Python 2 and 3.
# By Al Sweigart al@inventwithpython.com

__version__ = '0.1.29'

import collections
import datetime
import functools
import os
import subprocess
import sys
import time
import errno

from contextlib import contextmanager

from PIL import Image
from PIL import ImageOps
from PIL import ImageDraw
from PIL import __version__ as PIL__version__
from PIL import ImageGrab

useOpenCV: bool = False
try:
    import cv2
    import numpy

    useOpenCV = True
except ImportError:
    pass  # This is fine, useOpenCV will stay as False.

RUNNING_PYTHON_2 = sys.version_info[0] == 2

if not RUNNING_PYTHON_2:
    unicode = str  # On Python 3, all the isinstance(spam, (str, unicode)) calls will work the same as Python 2.

_PYGETWINDOW_UNAVAILABLE = True
if sys.platform == 'win32':
    # On Windows, the monitor scaling can be set to something besides normal 100%.
    # PyScreeze and Pillow needs to account for this to make accurate screenshots.
    # TODO - How does macOS and Linux handle monitor scaling?
    import ctypes

    try:
        ctypes.windll.user32.SetProcessDPIAware()
    except AttributeError:
        pass  # Windows XP doesn't support monitor scaling, so just do nothing.

    try:
        import pygetwindow
    except ImportError:
        _PYGETWINDOW_UNAVAILABLE = True
    else:
        _PYGETWINDOW_UNAVAILABLE = False
    


GRAYSCALE_DEFAULT = True

# For version 0.1.19 I changed it so that ImageNotFoundException was raised
# instead of returning None. In hindsight, this change came too late, so I'm
# changing it back to returning None. But I'm also including this option for
# folks who would rather have it raise an exception.
USE_IMAGE_NOT_FOUND_EXCEPTION = False

scrotExists = False
try:
    if sys.platform not in ('java', 'darwin', 'win32'):
        whichProc = subprocess.Popen(['which', 'scrot'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        scrotExists = whichProc.wait() == 0
except OSError as ex:
    if ex.errno == errno.ENOENT:
        # if there is no "which" program to find scrot, then assume there
        # is no scrot.
        pass
    else:
        raise


if sys.platform == 'win32':
    from ctypes import windll

    # win32 DC(DeviceContext) Manager
    @contextmanager
    def __win32_openDC(hWnd=0):
        """
        A context manager for handling calling GetDC() and ReleaseDC().

        This is used for win32 API calls, used by the pixel() function
        on Windows.

        Args:
            hWnd (int): The handle for the window to get a device context
        of, defaults to 0
        """
        hDC = windll.user32.GetDC(hWnd)
        if hDC == 0:  # NULL
            raise WindowsError("windll.user32.GetDC failed : return NULL")
        try:
            yield hDC
        finally:
            windll.user32.ReleaseDC.argtypes = [ctypes.c_ssize_t, ctypes.c_ssize_t]
            if windll.user32.ReleaseDC(hWnd, hDC) == 0:
                raise WindowsError("windll.user32.ReleaseDC failed : return 0")


Box = collections.namedtuple('Box', 'left top width height')
Point = collections.namedtuple('Point', 'x y')
RGB = collections.namedtuple('RGB', 'red green blue')


class PyScreezeException(Exception):
    """PyScreezeException is a generic exception class raised when a
    PyScreeze-related error happens. If a PyScreeze function raises an
    exception that isn't PyScreezeException or a subclass, assume it is
    a bug in PyScreeze."""

    pass


class ImageNotFoundException(PyScreezeException):
    """ImageNotFoundException is an exception class raised when the
    locate functions fail to locate an image. You must set
    pyscreeze.USE_IMAGE_NOT_FOUND_EXCEPTION to True to enable this feature.
    Otherwise, the locate functions will return None."""

    pass


def requiresPyGetWindow(wrappedFunction):
    """
    A decorator that marks a function as requiring PyGetWindow to be installed.
    This raises PyScreezeException if Pillow wasn't imported.
    """

    @functools.wraps(wrappedFunction)
    def wrapper(*args, **kwargs):
        if _PYGETWINDOW_UNAVAILABLE:
            raise PyScreezeException('The PyGetWindow package is required to use this function.')
        return wrappedFunction(*args, **kwargs)

    return wrapper


def _load_cv2(img, grayscale=None):
    """
    TODO
    """
    # load images if given filename, or convert as needed to opencv
    # Alpha layer just causes failures at this point, so flatten to RGB.
    # RGBA: load with -1 * cv2.CV_LOAD_IMAGE_COLOR to preserve alpha
    # to matchTemplate, need template and image to be the same wrt having alpha

    if grayscale is None:
        grayscale = GRAYSCALE_DEFAULT
    if isinstance(img, (str, unicode)):
        # The function imread loads an image from the specified file and
        # returns it. If the image cannot be read (because of missing
        # file, improper permissions, unsupported or invalid format),
        # the function returns an empty matrix
        # http://docs.opencv.org/3.0-beta/modules/imgcodecs/doc/reading_and_writing_images.html
        if grayscale:
            img_cv = cv2.imread(img, cv2.IMREAD_GRAYSCALE)
        else:
            img_cv = cv2.imread(img, cv2.IMREAD_COLOR)
        if img_cv is None:
            raise IOError(
                "Failed to read %s because file is missing, "
                "has improper permissions, or is an "
                "unsupported or invalid format" % img
            )
    elif isinstance(img, numpy.ndarray):
        # don't try to convert an already-gray image to gray
        if grayscale and len(img.shape) == 3:  # and img.shape[2] == 3:
            img_cv = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        else:
            img_cv = img
    elif hasattr(img, 'convert'):
        # assume its a PIL.Image, convert to cv format
        img_array = numpy.array(img.convert('RGB'))
        img_cv = img_array[:, :, ::-1].copy()  # -1 does RGB -> BGR
        if grayscale:
            img_cv = cv2.cvtColor(img_cv, cv2.COLOR_BGR2GRAY)
    else:
        raise TypeError('expected an image filename, OpenCV numpy array, or PIL image')
    return img_cv


def _locateAll_opencv(needleImage, haystackImage, grayscale=None, limit=10000, region=None, step=1, confidence=0.999):
    """
    TODO - rewrite this
        faster but more memory-intensive than pure python
        step 2 skips every other row and column = ~3x faster but prone to miss;
            to compensate, the algorithm automatically reduces the confidence
            threshold by 5% (which helps but will not avoid all misses).
        limitations:
          - OpenCV 3.x & python 3.x not tested
          - RGBA images are treated as RBG (ignores alpha channel)
    """
    if grayscale is None:
        grayscale = GRAYSCALE_DEFAULT

    confidence = float(confidence)

    needleImage = _load_cv2(needleImage, grayscale)
    needleHeight, needleWidth = needleImage.shape[:2]
    haystackImage = _load_cv2(haystackImage, grayscale)

    if region:
        haystackImage = haystackImage[region[1] : region[1] + region[3], region[0] : region[0] + region[2]]
    else:
        region = (0, 0)  # full image; these values used in the yield statement
    if haystackImage.shape[0] < needleImage.shape[0] or haystackImage.shape[1] < needleImage.shape[1]:
        # avoid semi-cryptic OpenCV error below if bad size
        raise ValueError('needle dimension(s) exceed the haystack image or region dimensions')

    if step == 2:
        confidence *= 0.95
        needleImage = needleImage[::step, ::step]
        haystackImage = haystackImage[::step, ::step]
    else:
        step = 1

    # get all matches at once, credit: https://stackoverflow.com/questions/7670112/finding-a-subimage-inside-a-numpy-image/9253805#9253805
    result = cv2.matchTemplate(haystackImage, needleImage, cv2.TM_CCOEFF_NORMED)
    match_indices = numpy.arange(result.size)[(result > confidence).flatten()]
    matches = numpy.unravel_index(match_indices[:limit], result.shape)

    if len(matches[0]) == 0:
        if USE_IMAGE_NOT_FOUND_EXCEPTION:
            raise ImageNotFoundException('Could not locate the image (highest confidence = %.3f)' % result.max())
        else:
            return

    # use a generator for API consistency:
    matchx = matches[1] * step + region[0]  # vectorized
    matchy = matches[0] * step + region[1]
    for x, y in zip(matchx, matchy):
        yield Box(x, y, needleWidth, needleHeight)


def _locateAll_pillow(needleImage, haystackImage, grayscale=None, limit=None, region=None, step=1, confidence=None):
    """
    TODO
    """
    if confidence is not None:
        raise NotImplementedError('The confidence keyword argument is only available if OpenCV is installed.')

    # setup all the arguments
    if grayscale is None:
        grayscale = GRAYSCALE_DEFAULT

    needleFileObj = None
    if isinstance(needleImage, (str, unicode)):
        # 'image' is a filename, load the Image object
        needleFileObj = open(needleImage, 'rb')
        needleImage = Image.open(needleFileObj)

    haystackFileObj = None
    if isinstance(haystackImage, (str, unicode)):
        # 'image' is a filename, load the Image object
        haystackFileObj = open(haystackImage, 'rb')
        haystackImage = Image.open(haystackFileObj)

    if region is not None:
        haystackImage = haystackImage.crop((region[0], region[1], region[0] + region[2], region[1] + region[3]))
    else:
        region = (0, 0)  # set to 0 because the code always accounts for a region

    if grayscale:  # if grayscale mode is on, convert the needle and haystack images to grayscale
        needleImage = ImageOps.grayscale(needleImage)
        haystackImage = ImageOps.grayscale(haystackImage)
    else:
        # if not using grayscale, make sure we are comparing RGB images, not RGBA images.
        if needleImage.mode == 'RGBA':
            needleImage = needleImage.convert('RGB')
        if haystackImage.mode == 'RGBA':
            haystackImage = haystackImage.convert('RGB')

    # setup some constants we'll be using in this function
    needleWidth, needleHeight = needleImage.size
    haystackWidth, haystackHeight = haystackImage.size

    needleImageData = tuple(needleImage.getdata())
    haystackImageData = tuple(haystackImage.getdata())

    needleImageRows = [
        needleImageData[y * needleWidth : (y + 1) * needleWidth] for y in range(needleHeight)
    ]  # LEFT OFF - check this
    needleImageFirstRow = needleImageRows[0]

    assert (
        len(needleImageFirstRow) == needleWidth
    ), 'For some reason, the calculated width of first row of the needle image is not the same as the width of the image.'
    assert [len(row) for row in needleImageRows] == [
        needleWidth
    ] * needleHeight, 'For some reason, the needleImageRows aren\'t the same size as the original image.'

    numMatchesFound = 0

    # NOTE: After running tests/benchmarks.py on the following code, it seem that having a step
    # value greater than 1 does not give *any* significant performance improvements.
    # Since using a step higher than 1 makes for less accurate matches, it will be
    # set to 1.
    step = 1  # hard-code step as 1 until a way to improve it can be figured out.

    if step == 1:
        firstFindFunc = _kmp
    else:
        firstFindFunc = _steppingFind

    for y in range(haystackHeight):  # start at the leftmost column
        for matchx in firstFindFunc(
            needleImageFirstRow, haystackImageData[y * haystackWidth : (y + 1) * haystackWidth], step
        ):
            foundMatch = True
            for searchy in range(1, needleHeight, step):
                haystackStart = (searchy + y) * haystackWidth + matchx
                if (
                    needleImageData[searchy * needleWidth : (searchy + 1) * needleWidth]
                    != haystackImageData[haystackStart : haystackStart + needleWidth]
                ):
                    foundMatch = False
                    break
            if foundMatch:
                # Match found, report the x, y, width, height of where the matching region is in haystack.
                numMatchesFound += 1
                yield Box(matchx + region[0], y + region[1], needleWidth, needleHeight)
                if limit is not None and numMatchesFound >= limit:
                    # Limit has been reached. Close file handles.
                    if needleFileObj is not None:
                        needleFileObj.close()
                    if haystackFileObj is not None:
                        haystackFileObj.close()
                    return

    # There was no limit or the limit wasn't reached, but close the file handles anyway.
    if needleFileObj is not None:
        needleFileObj.close()
    if haystackFileObj is not None:
        haystackFileObj.close()

    if numMatchesFound == 0:
        if USE_IMAGE_NOT_FOUND_EXCEPTION:
            raise ImageNotFoundException('Could not locate the image.')
        else:
            return


def locate(needleImage, haystackImage, **kwargs):
    """
    TODO
    """
    # Note: The gymnastics in this function is because we want to make sure to exhaust the iterator so that the needle and haystack files are closed in locateAll.
    kwargs['limit'] = 1
    points = tuple(locateAll(needleImage, haystackImage, **kwargs))
    if len(points) > 0:
        return points[0]
    else:
        if USE_IMAGE_NOT_FOUND_EXCEPTION:
            raise ImageNotFoundException('Could not locate the image.')
        else:
            return None


def locateOnScreen(image, minSearchTime=0, **kwargs):
    """TODO - rewrite this
    minSearchTime - amount of time in seconds to repeat taking
    screenshots and trying to locate a match.  The default of 0 performs
    a single search.
    """
    start = time.time()
    while True:
        try:
            screenshotIm = screenshot(
                region=None
            )  # the locateAll() function must handle cropping to return accurate coordinates, so don't pass a region here.
            retVal = locate(image, screenshotIm, **kwargs)
            try:
                screenshotIm.fp.close()
            except AttributeError:
                # Screenshots on Windows won't have an fp since they came from
                # ImageGrab, not a file. Screenshots on Linux will have fp set
                # to None since the file has been unlinked
                pass
            if retVal or time.time() - start > minSearchTime:
                return retVal
        except ImageNotFoundException:
            if time.time() - start > minSearchTime:
                if USE_IMAGE_NOT_FOUND_EXCEPTION:
                    raise
                else:
                    return None


def locateAllOnScreen(image, **kwargs):
    """
    TODO
    """

    # TODO - Should this raise an exception if zero instances of the image can be found on the screen, instead of always returning a generator?
    screenshotIm = screenshot(
        region=None
    )  # the locateAll() function must handle cropping to return accurate coordinates, so don't pass a region here.
    retVal = locateAll(image, screenshotIm, **kwargs)
    try:
        screenshotIm.fp.close()
    except AttributeError:
        # Screenshots on Windows won't have an fp since they came from
        # ImageGrab, not a file. Screenshots on Linux will have fp set
        # to None since the file has been unlinked
        pass
    return retVal


def locateCenterOnScreen(image, **kwargs):
    """
    TODO
    """
    coords = locateOnScreen(image, **kwargs)
    if coords is None:
        return None
    else:
        return center(coords)


def locateOnScreenNear(image, x, y):
    """
    TODO
    """

    foundMatchesBoxes = list(locateAllOnScreen(image))

    # NOTE: distances actually contains the distances squared, since getting the square root is expensive and unnecessary
    distances = []  # images[i] is related to distances[i]
    shortestDistanceIndex = 0  # The index of the shortest distance in `distances`

    # getting distance of all points from given point
    for foundMatchesBox in foundMatchesBoxes:
        foundMatchX, foundMatchY = center(foundMatchesBox)
        xDistance = abs(x - foundMatchX)
        yDistance = abs(y - foundMatchY)
        distances.append(xDistance * xDistance + yDistance * yDistance)

        if distances[-1] < distances[shortestDistanceIndex]:
            shortestDistanceIndex = len(distances) - 1

    # Returns the Box object of the match closest to x, y
    return foundMatchesBoxes[shortestDistanceIndex]


def locateCenterOnScreenNear(image, x, y, **kwargs):
    """
    TODO
    """
    coords = locateOnScreenNear(image, x, y, **kwargs)
    if coords is None:
        return None
    else:
        return center(coords)


@requiresPyGetWindow
def locateOnWindow(image, title, **kwargs):
    """
    TODO
    """
    matchingWindows = pygetwindow.getWindowsWithTitle(title)
    if len(matchingWindows) == 0:
        raise PyScreezeException('Could not find a window with %s in the title' % (title))
    elif len(matchingWindows) > 1:
        raise PyScreezeException(
            'Found multiple windows with %s in the title: %s' % (title, [str(win) for win in matchingWindows])
        )

    win = matchingWindows[0]
    win.activate()
    return locateOnScreen(image, region=(win.left, win.top, win.width, win.height), **kwargs)


@requiresPyGetWindow
def screenshotWindow(title):
    """
    TODO
    """
    pass  # Not implemented yet.


def showRegionOnScreen(region, outlineColor='red', filename='_showRegionOnScreen.png'):
    """
    TODO
    """
    # TODO - This function is useful! Document it!
    screenshotIm = screenshot()
    draw = ImageDraw.Draw(screenshotIm)
    region = (
        region[0],
        region[1],
        region[2] + region[0],
        region[3] + region[1],
    )  # convert from (left, top, right, bottom) to (left, top, width, height)
    draw.rectangle(region, outline=outlineColor)
    screenshotIm.save(filename)


def _screenshot_win32(imageFilename=None, region=None):
    """
    TODO
    """
    # TODO - Use the winapi to get a screenshot, and compare performance with ImageGrab.grab()
    # https://stackoverflow.com/a/3586280/1893164
    im = ImageGrab.grab()
    if region is not None:
        assert len(region) == 4, 'region argument must be a tuple of four ints'
        region = [int(x) for x in region]
        im = im.crop((region[0], region[1], region[2] + region[0], region[3] + region[1]))
    if imageFilename is not None:
        im.save(imageFilename)
    return im


def _screenshot_osx(imageFilename=None, region=None):
    """
    TODO
    """
    # TODO - use tmp name for this file.
    if tuple(PIL__version__) < (6, 2, 1):
        # Use the screencapture program if Pillow is older than 6.2.1, which
        # is when Pillow supported ImageGrab.grab() on macOS. (It may have
        # supported it earlier than 6.2.1, but I haven't tested it.)
        if imageFilename is None:
            tmpFilename = 'screenshot%s.png' % (datetime.datetime.now().strftime('%Y-%m%d_%H-%M-%S-%f'))
        else:
            tmpFilename = imageFilename
        subprocess.call(['screencapture', '-x', tmpFilename])
        im = Image.open(tmpFilename)

        if region is not None:
            assert len(region) == 4, 'region argument must be a tuple of four ints'
            region = [int(x) for x in region]
            im = im.crop((region[0], region[1], region[2] + region[0], region[3] + region[1]))
            os.unlink(tmpFilename)  # delete image of entire screen to save cropped version
            im.save(tmpFilename)
        else:
            # force loading before unlinking, Image.open() is lazy
            im.load()

        if imageFilename is None:
            os.unlink(tmpFilename)
    else:
        # Use ImageGrab.grab() to get the screenshot if Pillow version 6.3.2 or later is installed.
        im = ImageGrab.grab()
    return im


def _screenshot_linux(imageFilename=None, region=None):
    """
    TODO
    """

    # NOTE: scrot doesn't run correctly on Wayland, it only runs on x11.

    if not scrotExists:
        raise NotImplementedError(
            '"scrot" must be installed to use screenshot functions in Linux. Run: sudo apt-get install scrot'
        )
    if imageFilename is None:
        tmpFilename = '.screenshot%s.png' % (datetime.datetime.now().strftime('%Y-%m%d_%H-%M-%S-%f'))
    else:
        tmpFilename = imageFilename
    if scrotExists:
        subprocess.call(['scrot', '-z', tmpFilename])
        im = Image.open(tmpFilename)

        if region is not None:
            assert len(region) == 4, 'region argument must be a tuple of four ints'
            region = [int(x) for x in region]
            im = im.crop((region[0], region[1], region[2] + region[0], region[3] + region[1]))
            os.unlink(tmpFilename)  # delete image of entire screen to save cropped version
            im.save(tmpFilename)
        else:
            # force loading before unlinking, Image.open() is lazy
            im.load()

        if imageFilename is None:
            os.unlink(tmpFilename)
        return im
    else:
        raise Exception(
            'The scrot program must be installed to take a screenshot with PyScreeze on Linux. Run: sudo apt-get install scrot'
        )


def _kmp(needle, haystack, _dummy):  # Knuth-Morris-Pratt search algorithm implementation (to be used by screen capture)
    """
    TODO
    """
    # build table of shift amounts
    shifts = [1] * (len(needle) + 1)
    shift = 1
    for pos in range(len(needle)):
        while shift <= pos and needle[pos] != needle[pos - shift]:
            shift += shifts[pos - shift]
        shifts[pos + 1] = shift

    # do the actual search
    startPos = 0
    matchLen = 0
    for c in haystack:
        while matchLen == len(needle) or matchLen >= 0 and needle[matchLen] != c:
            startPos += shifts[matchLen]
            matchLen -= shifts[matchLen]
        matchLen += 1
        if matchLen == len(needle):
            yield startPos


def _steppingFind(needle, haystack, step):
    """
    TODO
    """
    for startPos in range(0, len(haystack) - len(needle) + 1):
        foundMatch = True
        for pos in range(0, len(needle), step):
            if haystack[startPos + pos] != needle[pos]:
                foundMatch = False
                break
        if foundMatch:
            yield startPos


def center(coords):
    """
    Returns a `Point` object with the x and y set to an integer determined by the format of `coords`.

    The `coords` argument is a 4-integer tuple of (left, top, width, height).

    For example:

    >>> center((10, 10, 6, 8))
    Point(x=13, y=14)
    >>> center((10, 10, 7, 9))
    Point(x=13, y=14)
    >>> center((10, 10, 8, 10))
    Point(x=14, y=15)
    """

    # TODO - one day, add code to handle a Box namedtuple.
    return Point(coords[0] + int(coords[2] / 2), coords[1] + int(coords[3] / 2))


def pixelMatchesColor(x, y, expectedRGBColor, tolerance=0):
    """
    TODO
    """
    pix = pixel(x, y)
    if len(pix) == 3 or len(expectedRGBColor) == 3:  # RGB mode
        r, g, b = pix[:3]
        exR, exG, exB = expectedRGBColor[:3]
        return (abs(r - exR) <= tolerance) and (abs(g - exG) <= tolerance) and (abs(b - exB) <= tolerance)
    elif len(pix) == 4 and len(expectedRGBColor) == 4:  # RGBA mode
        r, g, b, a = pix
        exR, exG, exB, exA = expectedRGBColor
        return (
            (abs(r - exR) <= tolerance)
            and (abs(g - exG) <= tolerance)
            and (abs(b - exB) <= tolerance)
            and (abs(a - exA) <= tolerance)
        )
    else:
        assert False, (
            'Color mode was expected to be length 3 (RGB) or 4 (RGBA), but pixel is length %s and expectedRGBColor is length %s'
            % (len(pix), len(expectedRGBColor))
        )


def pixel(x, y):
    """
    TODO
    """
    if sys.platform == 'win32':
        # On Windows, calling GetDC() and GetPixel() is twice as fast as using our screenshot() function.
        with __win32_openDC() as hdc:  # handle will be released automatically
            color = windll.gdi32.GetPixel(hdc, x, y)
            if color < 0:
                raise WindowsError("windll.gdi32.GetPixel failed : return {}".format(color))
            # color is in the format 0xbbggrr https://msdn.microsoft.com/en-us/library/windows/desktop/dd183449(v=vs.85).aspx
            bbggrr = "{:0>6x}".format(color)  # bbggrr => 'bbggrr' (hex)
            b, g, r = (int(bbggrr[i : i + 2], 16) for i in range(0, 6, 2))
            return (r, g, b)
    else:
        # Need to select only the first three values of the color in
        # case the returned pixel has an alpha channel
        return RGB(*(screenshot().getpixel((x, y))[:3]))


# set the screenshot() function based on the platform running this module
if sys.platform.startswith('java'):
    # Realistically, Jython will never be supported by PyScreeze.
    raise NotImplementedError('Jython is not yet supported by PyScreeze.')
elif sys.platform == 'darwin':
    screenshot = _screenshot_osx
elif sys.platform == 'win32':
    screenshot = _screenshot_win32
else:  
    # Everything else is considered to be Linux.
    screenshot = _screenshot_linux


# set the locateAll function to use opencv if possible; python 3 needs opencv 3.0+
# TODO - Should this raise an exception if zero instances of the image can be found on the screen, instead of always returning a generator?
locateAll = _locateAll_pillow
if useOpenCV:
    locateAll = _locateAll_opencv
    if not RUNNING_PYTHON_2 and cv2.__version__ < '3':
        locateAll = _locateAll_pillow




from pynput.keyboard import Key

from GameBotHelper import GameBotHelper


def do_stuff():
    print("Doing stuff")


if __name__ == '__main__':
    helper = GameBotHelper()
    helper.add_runnable(Key.f1, do_stuff)

    helper.start()



import logging

from PIL import Image

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.util import platform_is_osx

log = logging.getLogger(__name__)

# based on:
# http://stackoverflow.com/questions/69645/take-a-screenshot-via-a-python-script-linux


class WxBackendError(Exception):
    pass


app = None


class WxScreen(CBackend):
    name = "wx"
    # conflict with pygdk3
    #  wx is never installed by default
    # pygdk3 is default on Gnome

    def grab(self, bbox=None):
        if platform_is_osx():
            raise WxBackendError("osx not supported")
        import wx  # type: ignore

        global app
        if not app:
            app = wx.App()
        screen = wx.ScreenDC()
        size = screen.GetSize()
        if wx.__version__ >= "4":
            bmp = wx.Bitmap(size[0], size[1])
        else:
            bmp = wx.EmptyBitmap(size[0], size[1])
        mem = wx.MemoryDC(bmp)
        mem.Blit(0, 0, size[0], size[1], screen, 0, 0)
        del mem
        if hasattr(bmp, "ConvertToImage"):
            myWxImage = bmp.ConvertToImage()
        else:
            myWxImage = wx.ImageFromBitmap(bmp)
        im = Image.new("RGB", (myWxImage.GetWidth(), myWxImage.GetHeight()))
        if hasattr(Image, "frombytes"):
            # for Pillow
            im.frombytes(bytes(myWxImage.GetData()))
        else:
            # for PIL
            im.fromstring(myWxImage.GetData())
        if bbox:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        import wx

        return wx.__version__



import cv2
import numpy as np
import pyautogui
import random
import time
import platform
import subprocess
import os
import mss

is_retina = False
if platform.system() == "Darwin":
    is_retina = subprocess.call("system_profiler SPDisplaysDataType | grep -i 'retina'", shell=True) == 0

'''

grabs a region (topx, topy, bottomx, bottomy)
to the tuple (topx, topy, width, height)

input : a tuple containing the 4 coordinates of the region to capture

output : a PIL image of the area selected.

'''


def region_grabber(region):
    if is_retina: region = [n * 2 for n in region]
    x1 = region[0]
    y1 = region[1]
    width = region[2]
    height = region[3]

    region = x1, y1, width, height
    with mss.mss() as sct:
        return sct.grab(region)

'''

Searchs for an image within an area

input :

image : path to the image file (see opencv imread for supported types)
x1 : top left x value
y1 : top left y value
x2 : bottom right x value
y2 : bottom right y value
precision : the higher, the lesser tolerant and fewer false positives are found default is 0.8
im : a PIL image, usefull if you intend to search the same unchanging region for several elements

returns :
the top left corner coordinates of the element if found as an array [x,y] or [-1,-1] if not

'''


def imagesearcharea(image, x1, y1, x2, y2, precision=0.8, im=None):
    if im is None:
        im = region_grabber(region=(x1, y1, x2, y2))
        if is_retina:
            im.thumbnail((round(im.size[0] * 0.5), round(im.size[1] * 0.5)))
        # im.save('testarea.png') usefull for debugging purposes, this will save the captured region as "testarea.png"

    img_rgb = np.array(im)
    img_gray = cv2.cvtColor(img_rgb, cv2.COLOR_BGR2GRAY)
    template = cv2.imread(image, 0)
    if template is None:
        raise FileNotFoundError('Image file not found: {}'.format(image))

    res = cv2.matchTemplate(img_gray, template, cv2.TM_CCOEFF_NORMED)
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
    if max_val < precision:
        return [-1, -1]
    return max_loc


'''

click on the center of an image with a bit of random.
eg, if an image is 100*100 with an offset of 5 it may click at 52,50 the first time and then 55,53 etc
Usefull to avoid anti-bot monitoring while staying precise.

this function doesn't search for the image, it's only ment for easy clicking on the images.

input :

image : path to the image file (see opencv imread for supported types)
pos : array containing the position of the top left corner of the image [x,y]
action : button of the mouse to activate : "left" "right" "middle", see pyautogui.click documentation for more info
time : time taken for the mouse to move from where it was to the new position
'''


def click_image(image, pos, action, timestamp, offset=5):
    img = cv2.imread(image)
    if img is None:
        raise FileNotFoundError('Image file not found: {}'.format(image))
    height, width, channels = img.shape
    pyautogui.moveTo(pos[0] + r(width / 2, offset), pos[1] + r(height / 2, offset),
                     timestamp)
    pyautogui.click(button=action)


'''
Searchs for an image on the screen

input :

image : path to the image file (see opencv imread for supported types)
precision : the higher, the lesser tolerant and fewer false positives are found default is 0.8
im : a PIL image, usefull if you intend to search the same unchanging region for several elements

returns :
the top left corner coordinates of the element if found as an array [x,y] or [-1,-1] if not

'''


def imagesearch(image, precision=0.8):
    with mss.mss() as sct:
        im = sct.grab(sct.monitors[0])
        if is_retina:
            im.thumbnail((round(im.size[0] * 0.5), round(im.size[1] * 0.5)))
        # im.save('testarea.png') useful for debugging purposes, this will save the captured region as "testarea.png"
        img_rgb = np.array(im)
        img_gray = cv2.cvtColor(img_rgb, cv2.COLOR_BGR2GRAY)
        template = cv2.imread(image, 0)
        if template is None:
            raise FileNotFoundError('Image file not found: {}'.format(image))
        template.shape[::-1]

        res = cv2.matchTemplate(img_gray, template, cv2.TM_CCOEFF_NORMED)
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
        if max_val < precision:
            return [-1, -1]
        return max_loc


'''
Searchs for an image on screen continuously until it's found.

input :
image : path to the image file (see opencv imread for supported types)
time : Waiting time after failing to find the image
precision : the higher, the lesser tolerant and fewer false positives are found default is 0.8

returns :
the top left corner coordinates of the element if found as an array [x,y]

'''


def imagesearch_loop(image, timesample, precision=0.8):
    pos = imagesearch(image, precision)
    while pos[0] == -1:
        print(image + " not found, waiting")
        time.sleep(timesample)
        pos = imagesearch(image, precision)
    return pos


'''
Searchs for an image on screen continuously until it's found or max number of samples reached.

input :
image : path to the image file (see opencv imread for supported types)
time : Waiting time after failing to find the image
maxSamples: maximum number of samples before function times out.
precision : the higher, the lesser tolerant and fewer false positives are found default is 0.8

returns :
the top left corner coordinates of the element if found as an array [x,y]

'''


def imagesearch_numLoop(image, timesample, maxSamples, precision=0.8):
    pos = imagesearch(image, precision)
    count = 0
    while pos[0] == -1:
        print(image + " not found, waiting")
        time.sleep(timesample)
        pos = imagesearch(image, precision)
        count = count + 1
        if count > maxSamples:
            break
    return pos


'''
Searchs for an image on a region of the screen continuously until it's found.

input :
image : path to the image file (see opencv imread for supported types)
time : Waiting time after failing to find the image
x1 : top left x value
y1 : top left y value
x2 : bottom right x value
y2 : bottom right y value
precision : the higher, the lesser tolerant and fewer false positives are found default is 0.8

returns :
the top left corner coordinates of the element as an array [x,y]

'''


def imagesearch_region_loop(image, timesample, x1, y1, x2, y2, precision=0.8):
    pos = imagesearcharea(image, x1, y1, x2, y2, precision)

    while pos[0] == -1:
        time.sleep(timesample)
        pos = imagesearcharea(image, x1, y1, x2, y2, precision)
    return pos


'''
Searchs for an image on a region of the screen continuously until it's found or max number of samples reached.

input :
image : path to the image file (see opencv imread for supported types)
time : Waiting time after failing to find the image
maxSamples: maximum number of samples before function times out.
x1 : top left x value
y1 : top left y value
x2 : bottom right x value
y2 : bottom right y value
precision : the higher, the lesser tolerant and fewer false positives are found default is 0.8

returns :
the top left corner coordinates of the element as an array [x,y]
'''


def imagesearch_region_numLoop(image, timesample, maxSamples, x1, y1, x2, y2, precision=0.8):
    pos = imagesearcharea(image, x1, y1, x2, y2, precision)
    count = 0
    while pos[0] == -1:
        time.sleep(timesample)
        pos = imagesearcharea(image, x1, y1, x2, y2, precision)
        count = count + 1
        if count > maxSamples:
            break
    return pos



'''
Searches for an image on the screen and counts the number of occurrences.

input :
image : path to the target image file (see opencv imread for supported types)
precision : the higher, the lesser tolerant and fewer false positives are found default is 0.9

returns :
the number of times a given image appears on the screen.
optionally an output image with all the occurances boxed with a red outline.

'''


def imagesearch_count(image, precision=0.95):
    with mss.mss() as sct:
        im = sct.grab(sct.monitors[0])
        if is_retina:
            im.thumbnail((round(im.size[0] * 0.5), round(im.size[1] * 0.5)))
        img_rgb = np.array(im)
        img_gray = cv2.cvtColor(img_rgb, cv2.COLOR_BGR2GRAY)
        template = cv2.imread(image, 0)
        if template is None:
            raise FileNotFoundError('Image file not found: {}'.format(image))
        template.shape[::-1]
        res = cv2.matchTemplate(img_gray, template, cv2.TM_CCOEFF_NORMED)
        loc = np.where(res >= precision)
        count = 0
        for pt in zip(*loc[::-1]):  # Swap columns and rows
            count = count + 1
        return count


'''
Get all screens on the provided folder and search them on screen.

input :
path : path of the folder with the images to be searched on screen
precision : the higher, the lesser tolerant and fewer false positives are found default is 0.8

returns :
A dictionary where the key is the path to image file and the value is the position where was found.
'''


def imagesearch_from_folder(path, precision):
    print(path)
    imagesPos = {}
    path = path if path[-1] == '/' or '\\' else path+'/'
    valid_images = [".jpg", ".gif", ".png", ".jpeg"]
    files = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f)) and os.path.splitext(f)[1].lower() in valid_images]
    for file in files:
        pos = imagesearch(path+file, precision)
        imagesPos[path+file] = pos
    return imagesPos


def r(num, rand):
    return num + rand * random.random()

'''
Wrapper around imagesearch and click_image

# TODO: optimize so that we only read the file once.

input :
image : path to the image file (see opencv imread for supported types)
action : button of the mouse to activate : "left" "right" "middle", see pyautogui.click documentation for more info
delay : time taken for the mouse to move from where it was to the new position
precision : the higher, the lesser tolerant and fewer false positives are found default is 0.8
'''

def imagesearch_click(image, action, delay, offset=5, precision=0.8):
    pos = imagesearch(image, precision)
    img = cv2.imread(image)
    if img is None:
        raise FileNotFoundError('Image file not found: {}'.format(image))
    height, width, channels = img.shape
    pyautogui.moveTo(pos[0] + r(width / 2, offset), pos[1] + r(height / 2, offset),
                     delay)
    pyautogui.click(button=action)



__all__ = [
    "IntPointer",
    "MatLike",
    "MatShape",
    "Size",
    "Size2f",
    "Scalar",
    "Point",
    "Point2i",
    "Point2f",
    "Point2d",
    "Point3i",
    "Point3f",
    "Point3d",
    "Range",
    "Rect",
    "Rect2i",
    "Rect2d",
    "Moments",
    "RotatedRect",
    "TermCriteria",
    "Vec2i",
    "Vec2f",
    "Vec2d",
    "Vec3i",
    "Vec3f",
    "Vec3d",
    "Vec4i",
    "Vec4f",
    "Vec4d",
    "Vec6f",
    "FeatureDetector",
    "DescriptorExtractor",
    "FeatureExtractor",
    "GProtoArg",
    "GProtoInputArgs",
    "GProtoOutputArgs",
    "GRunArg",
    "GOptRunArg",
    "GMetaArg",
    "Prim",
    "Matx33f",
    "Matx33d",
    "Matx44f",
    "Matx44d",
    "GTypeInfo",
    "ExtractArgsCallback",
    "ExtractMetaCallback",
    "LayerId",
    "IndexParams",
    "SearchParams",
    "map_string_and_string",
    "map_string_and_int",
    "map_string_and_vector_size_t",
    "map_string_and_vector_float",
    "map_int_and_double",
]

import cv2
import cv2.dnn
import cv2.gapi.wip.draw
import cv2.mat_wrapper
import numpy
import sys
import typing


if numpy.lib.NumpyVersion(numpy.__version__) > "1.20.0" and sys.version_info >= (3, 9):
    NumPyArrayGeneric = numpy.ndarray[typing.Any, numpy.dtype[numpy.generic]]
else:
    NumPyArrayGeneric = numpy.ndarray


if numpy.lib.NumpyVersion(numpy.__version__) > "1.20.0" and sys.version_info >= (3, 9):
    NumPyArrayFloat32 = numpy.ndarray[typing.Any, numpy.dtype[numpy.float32]]
else:
    NumPyArrayFloat32 = numpy.ndarray


if numpy.lib.NumpyVersion(numpy.__version__) > "1.20.0" and sys.version_info >= (3, 9):
    NumPyArrayFloat64 = numpy.ndarray[typing.Any, numpy.dtype[numpy.float64]]
else:
    NumPyArrayFloat64 = numpy.ndarray


if typing.TYPE_CHECKING:
    TermCriteria_Type = cv2.TermCriteria_Type
else:
    TermCriteria_Type = int


IntPointer = int
"""Represents an arbitrary pointer"""
MatLike = typing.Union[cv2.mat_wrapper.Mat, NumPyArrayGeneric]
MatShape = typing.Sequence[int]
Size = typing.Sequence[int]
"""Required length is 2"""
Size2f = typing.Sequence[float]
"""Required length is 2"""
Scalar = typing.Sequence[float]
"""Required length is at most 4"""
Point = typing.Sequence[int]
"""Required length is 2"""
Point2i = Point
Point2f = typing.Sequence[float]
"""Required length is 2"""
Point2d = typing.Sequence[float]
"""Required length is 2"""
Point3i = typing.Sequence[int]
"""Required length is 3"""
Point3f = typing.Sequence[float]
"""Required length is 3"""
Point3d = typing.Sequence[float]
"""Required length is 3"""
Range = typing.Sequence[int]
"""Required length is 2"""
Rect = typing.Sequence[int]
"""Required length is 4"""
Rect2i = typing.Sequence[int]
"""Required length is 4"""
Rect2d = typing.Sequence[float]
"""Required length is 4"""
Moments = typing.Dict[str, float]
RotatedRect = typing.Tuple[Point2f, Size, float]
"""Any type providing sequence protocol is supported"""
TermCriteria = typing.Tuple[TermCriteria_Type, int, float]
"""Any type providing sequence protocol is supported"""
Vec2i = typing.Sequence[int]
"""Required length is 2"""
Vec2f = typing.Sequence[float]
"""Required length is 2"""
Vec2d = typing.Sequence[float]
"""Required length is 2"""
Vec3i = typing.Sequence[int]
"""Required length is 3"""
Vec3f = typing.Sequence[float]
"""Required length is 3"""
Vec3d = typing.Sequence[float]
"""Required length is 3"""
Vec4i = typing.Sequence[int]
"""Required length is 4"""
Vec4f = typing.Sequence[float]
"""Required length is 4"""
Vec4d = typing.Sequence[float]
"""Required length is 4"""
Vec6f = typing.Sequence[float]
"""Required length is 6"""
FeatureDetector = cv2.Feature2D
DescriptorExtractor = cv2.Feature2D
FeatureExtractor = cv2.Feature2D
GProtoArg = typing.Union[Scalar, cv2.GMat, cv2.GOpaqueT, cv2.GArrayT]
GProtoInputArgs = typing.Sequence[GProtoArg]
GProtoOutputArgs = typing.Sequence[GProtoArg]
GRunArg = typing.Union[MatLike, Scalar, cv2.GOpaqueT, cv2.GArrayT, typing.Sequence[typing.Any], None]
GOptRunArg = typing.Optional[GRunArg]
GMetaArg = typing.Union[cv2.GMat, Scalar, cv2.GOpaqueT, cv2.GArrayT]
Prim = typing.Union[cv2.gapi.wip.draw.Text, cv2.gapi.wip.draw.Circle, cv2.gapi.wip.draw.Image, cv2.gapi.wip.draw.Line, cv2.gapi.wip.draw.Rect, cv2.gapi.wip.draw.Mosaic, cv2.gapi.wip.draw.Poly]
Matx33f = NumPyArrayFloat32
"""NDArray(shape=(3, 3), dtype=numpy.float32)"""
Matx33d = NumPyArrayFloat64
"""NDArray(shape=(3, 3), dtype=numpy.float64)"""
Matx44f = NumPyArrayFloat32
"""NDArray(shape=(4, 4), dtype=numpy.float32)"""
Matx44d = NumPyArrayFloat64
"""NDArray(shape=(4, 4), dtype=numpy.float64)"""
GTypeInfo = typing.Union[cv2.GMat, Scalar, cv2.GOpaqueT, cv2.GArrayT]
ExtractArgsCallback = typing.Callable[[typing.Sequence[GTypeInfo]], typing.Sequence[GRunArg]]
ExtractMetaCallback = typing.Callable[[typing.Sequence[GTypeInfo]], typing.Sequence[GMetaArg]]
LayerId = cv2.dnn.DictValue
IndexParams = typing.Dict[str, typing.Union[bool, int, float, str]]
SearchParams = typing.Dict[str, typing.Union[bool, int, float, str]]
map_string_and_string = typing.Dict[str, str]
map_string_and_int = typing.Dict[str, int]
map_string_and_vector_size_t = typing.Dict[str, typing.Sequence[int]]
map_string_and_vector_float = typing.Dict[str, typing.Sequence[float]]
map_int_and_double = typing.Dict[int, float]



import io
import logging

from PIL import Image

from pyscreenshot.plugins.backend import CBackend

log = logging.getLogger(__name__)

# based on qt4 backend

app = None


class Qt5GrabWindow(CBackend):
    name = "pyqt5"

    # qt backends have conflict with each other in the same process.

    def grab_to_buffer(self, buff, file_type="png"):
        from PyQt5 import Qt, QtGui, QtWidgets  # type: ignore

        QApplication = QtWidgets.QApplication
        QBuffer = Qt.QBuffer
        QIODevice = Qt.QIODevice
        QScreen = QtGui.QScreen

        global app
        if not app:
            app = QApplication([])
        qbuffer = QBuffer()
        qbuffer.open(QIODevice.ReadWrite)
        QScreen.grabWindow(
            QApplication.primaryScreen(), QApplication.desktop().winId()
        ).save(qbuffer, file_type)
        buff.write(qbuffer.data())
        qbuffer.close()

    def grab(self, bbox=None):
        strio = io.BytesIO()
        self.grab_to_buffer(strio)
        strio.seek(0)
        im = Image.open(strio)
        if bbox:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        from PyQt5 import Qt

        return Qt.PYQT_VERSION_STR



from PIL import __version__

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.util import use_x_display


class PilWrapper(CBackend):
    name = "pil"

    def grab(self, bbox=None):
        from PIL import ImageGrab

        # https://pillow.readthedocs.io/en/stable/reference/ImageGrab.html
        # On Linux, if xdisplay is None then gnome-screenshot will be used if it is installed.
        # To capture the default X11 display instead, pass xdisplay=""
        xdisplay = None
        if use_x_display():
            xdisplay = ""
        return ImageGrab.grab(bbox, xdisplay=xdisplay)

    def backend_version(self):
        return __version__



import sys
from importlib.machinery import ModuleSpec, PathFinder
from importlib.machinery import all_suffixes as module_suffixes
from importlib.util import spec_from_file_location
from itertools import chain
from pathlib import Path

MAPPING = {'src': 'C:\\Users\\Chris\\repos\\GameBotHelper\\src', 'test': 'C:\\Users\\Chris\\repos\\GameBotHelper\\test', 'venv': 'C:\\Users\\Chris\\repos\\GameBotHelper\\venv'}
NAMESPACES = {'src': ['C:\\Users\\Chris\\repos\\GameBotHelper\\src'], 'test': ['C:\\Users\\Chris\\repos\\GameBotHelper\\test'], 'venv': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv'], 'venv.Lib': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib'], 'venv.Scripts': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Scripts'], 'venv.Lib.site-packages': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages'], 'venv.Lib.site-packages.jaraco': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\jaraco'], 'venv.Lib.site-packages.docutils.parsers.rst.include': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\docutils\\parsers\\rst\\include'], 'venv.Lib.site-packages.docutils.writers.s5_html.themes': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\docutils\\writers\\s5_html\\themes'], 'venv.Lib.site-packages.docutils.writers.s5_html.themes.big-black': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\docutils\\writers\\s5_html\\themes\\big-black'], 'venv.Lib.site-packages.docutils.writers.s5_html.themes.big-white': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\docutils\\writers\\s5_html\\themes\\big-white'], 'venv.Lib.site-packages.docutils.writers.s5_html.themes.default': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\docutils\\writers\\s5_html\\themes\\default'], 'venv.Lib.site-packages.docutils.writers.s5_html.themes.medium-black': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\docutils\\writers\\s5_html\\themes\\medium-black'], 'venv.Lib.site-packages.docutils.writers.s5_html.themes.medium-white': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\docutils\\writers\\s5_html\\themes\\medium-white'], 'venv.Lib.site-packages.docutils.writers.s5_html.themes.small-black': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\docutils\\writers\\s5_html\\themes\\small-black'], 'venv.Lib.site-packages.docutils.writers.s5_html.themes.small-white': ['C:\\Users\\Chris\\repos\\GameBotHelper\\venv\\Lib\\site-packages\\docutils\\writers\\s5_html\\themes\\small-white']}
PATH_PLACEHOLDER = '__editable__.game_bot_helper-0.0.1.finder' + ".__path_hook__"


class _EditableFinder:  # MetaPathFinder
    @classmethod
    def find_spec(cls, fullname, path=None, target=None):
        extra_path = []

        # Top-level packages and modules (we know these exist in the FS)
        if fullname in MAPPING:
            pkg_path = MAPPING[fullname]
            return cls._find_spec(fullname, Path(pkg_path))

        # Handle immediate children modules (required for namespaces to work)
        # To avoid problems with case sensitivity in the file system we delegate
        # to the importlib.machinery implementation.
        parent, _, child = fullname.rpartition(".")
        if parent and parent in MAPPING:
            return PathFinder.find_spec(fullname, path=[MAPPING[parent], *extra_path])

        # Other levels of nesting should be handled automatically by importlib
        # using the parent path.
        return None

    @classmethod
    def _find_spec(cls, fullname, candidate_path):
        init = candidate_path / "__init__.py"
        candidates = (candidate_path.with_suffix(x) for x in module_suffixes())
        for candidate in chain([init], candidates):
            if candidate.exists():
                return spec_from_file_location(fullname, candidate)


class _EditableNamespaceFinder:  # PathEntryFinder
    @classmethod
    def _path_hook(cls, path):
        if path == PATH_PLACEHOLDER:
            return cls
        raise ImportError

    @classmethod
    def _paths(cls, fullname):
        # Ensure __path__ is not empty for the spec to be considered a namespace.
        return NAMESPACES[fullname] or MAPPING.get(fullname) or [PATH_PLACEHOLDER]

    @classmethod
    def find_spec(cls, fullname, target=None):
        if fullname in NAMESPACES:
            spec = ModuleSpec(fullname, None, is_package=True)
            spec.submodule_search_locations = cls._paths(fullname)
            return spec
        return None

    @classmethod
    def find_module(cls, fullname):
        return None


def install():
    if not any(finder == _EditableFinder for finder in sys.meta_path):
        sys.meta_path.append(_EditableFinder)

    if not NAMESPACES:
        return

    if not any(hook == _EditableNamespaceFinder._path_hook for hook in sys.path_hooks):
        # PathEntryFinder is needed to create NamespaceSpec without private APIS
        sys.path_hooks.append(_EditableNamespaceFinder._path_hook)
    if PATH_PLACEHOLDER not in sys.path:
        sys.path.append(PATH_PLACEHOLDER)  # Used just to trigger the path hook



import cv2
import cv2.typing
import typing


# Enumerations
StereoOutputFormat_DEPTH_FLOAT16: int
STEREO_OUTPUT_FORMAT_DEPTH_FLOAT16: int
StereoOutputFormat_DEPTH_FLOAT32: int
STEREO_OUTPUT_FORMAT_DEPTH_FLOAT32: int
StereoOutputFormat_DISPARITY_FIXED16_11_5: int
STEREO_OUTPUT_FORMAT_DISPARITY_FIXED16_11_5: int
StereoOutputFormat_DISPARITY_FIXED16_12_4: int
STEREO_OUTPUT_FORMAT_DISPARITY_FIXED16_12_4: int
StereoOutputFormat_DEPTH_16F: int
STEREO_OUTPUT_FORMAT_DEPTH_16F: int
StereoOutputFormat_DEPTH_32F: int
STEREO_OUTPUT_FORMAT_DEPTH_32F: int
StereoOutputFormat_DISPARITY_16Q_10_5: int
STEREO_OUTPUT_FORMAT_DISPARITY_16Q_10_5: int
StereoOutputFormat_DISPARITY_16Q_11_4: int
STEREO_OUTPUT_FORMAT_DISPARITY_16Q_11_4: int
StereoOutputFormat = int
"""One of [StereoOutputFormat_DEPTH_FLOAT16, STEREO_OUTPUT_FORMAT_DEPTH_FLOAT16, StereoOutputFormat_DEPTH_FLOAT32, STEREO_OUTPUT_FORMAT_DEPTH_FLOAT32, StereoOutputFormat_DISPARITY_FIXED16_11_5, STEREO_OUTPUT_FORMAT_DISPARITY_FIXED16_11_5, StereoOutputFormat_DISPARITY_FIXED16_12_4, STEREO_OUTPUT_FORMAT_DISPARITY_FIXED16_12_4, StereoOutputFormat_DEPTH_16F, STEREO_OUTPUT_FORMAT_DEPTH_16F, StereoOutputFormat_DEPTH_32F, STEREO_OUTPUT_FORMAT_DEPTH_32F, StereoOutputFormat_DISPARITY_16Q_10_5, STEREO_OUTPUT_FORMAT_DISPARITY_16Q_10_5, StereoOutputFormat_DISPARITY_16Q_11_4, STEREO_OUTPUT_FORMAT_DISPARITY_16Q_11_4]"""

CV_BOOL: int
CV_INT: int
CV_INT64: int
CV_DOUBLE: int
CV_FLOAT: int
CV_STRING: int
CV_POINT: int
CV_POINT2F: int
CV_POINT3F: int
CV_SIZE: int
CV_RECT: int
CV_SCALAR: int
CV_MAT: int
CV_GMAT: int
CV_DRAW_PRIM: int
CV_ANY: int
ArgType = int
"""One of [CV_BOOL, CV_INT, CV_INT64, CV_DOUBLE, CV_FLOAT, CV_STRING, CV_POINT, CV_POINT2F, CV_POINT3F, CV_SIZE, CV_RECT, CV_SCALAR, CV_MAT, CV_GMAT, CV_DRAW_PRIM, CV_ANY]"""



# Classes
class GNetParam:
    ...

class GNetPackage:
    # Functions
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, nets: typing.Sequence[GNetParam]) -> None: ...



# Functions
def BGR2Gray(src: cv2.GMat) -> cv2.GMat: ...

def BGR2I420(src: cv2.GMat) -> cv2.GMat: ...

def BGR2LUV(src: cv2.GMat) -> cv2.GMat: ...

def BGR2RGB(src: cv2.GMat) -> cv2.GMat: ...

def BGR2YUV(src: cv2.GMat) -> cv2.GMat: ...

def BayerGR2RGB(src_gr: cv2.GMat) -> cv2.GMat: ...

def Canny(image: cv2.GMat, threshold1: float, threshold2: float, apertureSize: int = ..., L2gradient: bool = ...) -> cv2.GMat: ...

def I4202BGR(src: cv2.GMat) -> cv2.GMat: ...

def I4202RGB(src: cv2.GMat) -> cv2.GMat: ...

def LUT(src: cv2.GMat, lut: cv2.typing.MatLike) -> cv2.GMat: ...

def LUV2BGR(src: cv2.GMat) -> cv2.GMat: ...

def Laplacian(src: cv2.GMat, ddepth: int, ksize: int = ..., scale: float = ..., delta: float = ..., borderType: int = ...) -> cv2.GMat: ...

def NV12toBGR(src_y: cv2.GMat, src_uv: cv2.GMat) -> cv2.GMat: ...

def NV12toGray(src_y: cv2.GMat, src_uv: cv2.GMat) -> cv2.GMat: ...

def NV12toRGB(src_y: cv2.GMat, src_uv: cv2.GMat) -> cv2.GMat: ...

@typing.overload
def RGB2Gray(src: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def RGB2Gray(src: cv2.GMat, rY: float, gY: float, bY: float) -> cv2.GMat: ...

def RGB2HSV(src: cv2.GMat) -> cv2.GMat: ...

def RGB2I420(src: cv2.GMat) -> cv2.GMat: ...

def RGB2Lab(src: cv2.GMat) -> cv2.GMat: ...

def RGB2YUV(src: cv2.GMat) -> cv2.GMat: ...

def RGB2YUV422(src: cv2.GMat) -> cv2.GMat: ...

def Sobel(src: cv2.GMat, ddepth: int, dx: int, dy: int, ksize: int = ..., scale: float = ..., delta: float = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def SobelXY(src: cv2.GMat, ddepth: int, order: int, ksize: int = ..., scale: float = ..., delta: float = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> tuple[cv2.GMat, cv2.GMat]: ...

def YUV2BGR(src: cv2.GMat) -> cv2.GMat: ...

def YUV2RGB(src: cv2.GMat) -> cv2.GMat: ...

def absDiff(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...

def absDiffC(src: cv2.GMat, c: cv2.GScalar) -> cv2.GMat: ...

def add(src1: cv2.GMat, src2: cv2.GMat, ddepth: int = ...) -> cv2.GMat: ...

@typing.overload
def addC(src1: cv2.GMat, c: cv2.GScalar, ddepth: int = ...) -> cv2.GMat: ...
@typing.overload
def addC(c: cv2.GScalar, src1: cv2.GMat, ddepth: int = ...) -> cv2.GMat: ...

def addWeighted(src1: cv2.GMat, alpha: float, src2: cv2.GMat, beta: float, gamma: float, ddepth: int = ...) -> cv2.GMat: ...

def bilateralFilter(src: cv2.GMat, d: int, sigmaColor: float, sigmaSpace: float, borderType: int = ...) -> cv2.GMat: ...

@typing.overload
def bitwise_and(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def bitwise_and(src1: cv2.GMat, src2: cv2.GScalar) -> cv2.GMat: ...

def bitwise_not(src: cv2.GMat) -> cv2.GMat: ...

@typing.overload
def bitwise_or(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def bitwise_or(src1: cv2.GMat, src2: cv2.GScalar) -> cv2.GMat: ...

@typing.overload
def bitwise_xor(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def bitwise_xor(src1: cv2.GMat, src2: cv2.GScalar) -> cv2.GMat: ...

def blur(src: cv2.GMat, ksize: cv2.typing.Size, anchor: cv2.typing.Point = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

@typing.overload
def boundingRect(src: cv2.GMat) -> cv2.GOpaqueT: ...
@typing.overload
def boundingRect(src: cv2.GArrayT) -> cv2.GOpaqueT: ...
@typing.overload
def boundingRect(src: cv2.GArrayT) -> cv2.GOpaqueT: ...

def boxFilter(src: cv2.GMat, dtype: int, ksize: cv2.typing.Size, anchor: cv2.typing.Point = ..., normalize: bool = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def cartToPolar(x: cv2.GMat, y: cv2.GMat, angleInDegrees: bool = ...) -> tuple[cv2.GMat, cv2.GMat]: ...

@typing.overload
def cmpEQ(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def cmpEQ(src1: cv2.GMat, src2: cv2.GScalar) -> cv2.GMat: ...

@typing.overload
def cmpGE(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def cmpGE(src1: cv2.GMat, src2: cv2.GScalar) -> cv2.GMat: ...

@typing.overload
def cmpGT(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def cmpGT(src1: cv2.GMat, src2: cv2.GScalar) -> cv2.GMat: ...

@typing.overload
def cmpLE(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def cmpLE(src1: cv2.GMat, src2: cv2.GScalar) -> cv2.GMat: ...

@typing.overload
def cmpLT(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def cmpLT(src1: cv2.GMat, src2: cv2.GScalar) -> cv2.GMat: ...

@typing.overload
def cmpNE(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def cmpNE(src1: cv2.GMat, src2: cv2.GScalar) -> cv2.GMat: ...

def combine(lhs: cv2.GKernelPackage, rhs: cv2.GKernelPackage) -> cv2.GKernelPackage: ...

@typing.overload
def concatHor(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def concatHor(v: typing.Sequence[cv2.GMat]) -> cv2.GMat: ...

@typing.overload
def concatVert(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...
@typing.overload
def concatVert(v: typing.Sequence[cv2.GMat]) -> cv2.GMat: ...

def convertTo(src: cv2.GMat, rdepth: int, alpha: float = ..., beta: float = ...) -> cv2.GMat: ...

def copy(in_: cv2.GMat) -> cv2.GMat: ...

def countNonZero(src: cv2.GMat) -> cv2.GOpaqueT: ...

def crop(src: cv2.GMat, rect: cv2.typing.Rect) -> cv2.GMat: ...

def dilate(src: cv2.GMat, kernel: cv2.typing.MatLike, anchor: cv2.typing.Point = ..., iterations: int = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def dilate3x3(src: cv2.GMat, iterations: int = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def div(src1: cv2.GMat, src2: cv2.GMat, scale: float, ddepth: int = ...) -> cv2.GMat: ...

def divC(src: cv2.GMat, divisor: cv2.GScalar, scale: float, ddepth: int = ...) -> cv2.GMat: ...

def divRC(divident: cv2.GScalar, src: cv2.GMat, scale: float, ddepth: int = ...) -> cv2.GMat: ...

def equalizeHist(src: cv2.GMat) -> cv2.GMat: ...

def erode(src: cv2.GMat, kernel: cv2.typing.MatLike, anchor: cv2.typing.Point = ..., iterations: int = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def erode3x3(src: cv2.GMat, iterations: int = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def filter2D(src: cv2.GMat, ddepth: int, kernel: cv2.typing.MatLike, anchor: cv2.typing.Point = ..., delta: cv2.typing.Scalar = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def flip(src: cv2.GMat, flipCode: int) -> cv2.GMat: ...

def gaussianBlur(src: cv2.GMat, ksize: cv2.typing.Size, sigmaX: float, sigmaY: float = ..., borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def goodFeaturesToTrack(image: cv2.GMat, maxCorners: int, qualityLevel: float, minDistance: float, mask: cv2.typing.MatLike | None = ..., blockSize: int = ..., useHarrisDetector: bool = ..., k: float = ...) -> cv2.GArrayT: ...

def inRange(src: cv2.GMat, threshLow: cv2.GScalar, threshUp: cv2.GScalar) -> cv2.GMat: ...

@typing.overload
def infer(name: str, inputs: cv2.GInferInputs) -> cv2.GInferOutputs: ...
@typing.overload
def infer(name: str, roi: cv2.GOpaqueT, inputs: cv2.GInferInputs) -> cv2.GInferOutputs: ...
@typing.overload
def infer(name: str, rois: cv2.GArrayT, inputs: cv2.GInferInputs) -> cv2.GInferListOutputs: ...

def infer2(name: str, in_: cv2.GMat, inputs: cv2.GInferListInputs) -> cv2.GInferListOutputs: ...

def integral(src: cv2.GMat, sdepth: int = ..., sqdepth: int = ...) -> tuple[cv2.GMat, cv2.GMat]: ...

@typing.overload
def kmeans(data: cv2.GMat, K: int, bestLabels: cv2.GMat, criteria: cv2.typing.TermCriteria, attempts: int, flags: cv2.KmeansFlags) -> tuple[cv2.GOpaqueT, cv2.GMat, cv2.GMat]: ...
@typing.overload
def kmeans(data: cv2.GMat, K: int, criteria: cv2.typing.TermCriteria, attempts: int, flags: cv2.KmeansFlags) -> tuple[cv2.GOpaqueT, cv2.GMat, cv2.GMat]: ...
@typing.overload
def kmeans(data: cv2.GArrayT, K: int, bestLabels: cv2.GArrayT, criteria: cv2.typing.TermCriteria, attempts: int, flags: cv2.KmeansFlags) -> tuple[cv2.GOpaqueT, cv2.GArrayT, cv2.GArrayT]: ...
@typing.overload
def kmeans(data: cv2.GArrayT, K: int, bestLabels: cv2.GArrayT, criteria: cv2.typing.TermCriteria, attempts: int, flags: cv2.KmeansFlags) -> tuple[cv2.GOpaqueT, cv2.GArrayT, cv2.GArrayT]: ...

def mask(src: cv2.GMat, mask: cv2.GMat) -> cv2.GMat: ...

def max(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...

def mean(src: cv2.GMat) -> cv2.GScalar: ...

def medianBlur(src: cv2.GMat, ksize: int) -> cv2.GMat: ...

def merge3(src1: cv2.GMat, src2: cv2.GMat, src3: cv2.GMat) -> cv2.GMat: ...

def merge4(src1: cv2.GMat, src2: cv2.GMat, src3: cv2.GMat, src4: cv2.GMat) -> cv2.GMat: ...

def min(src1: cv2.GMat, src2: cv2.GMat) -> cv2.GMat: ...

def morphologyEx(src: cv2.GMat, op: cv2.MorphTypes, kernel: cv2.typing.MatLike, anchor: cv2.typing.Point = ..., iterations: int = ..., borderType: cv2.BorderTypes = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def mul(src1: cv2.GMat, src2: cv2.GMat, scale: float = ..., ddepth: int = ...) -> cv2.GMat: ...

@typing.overload
def mulC(src: cv2.GMat, multiplier: float, ddepth: int = ...) -> cv2.GMat: ...
@typing.overload
def mulC(src: cv2.GMat, multiplier: cv2.GScalar, ddepth: int = ...) -> cv2.GMat: ...
@typing.overload
def mulC(multiplier: cv2.GScalar, src: cv2.GMat, ddepth: int = ...) -> cv2.GMat: ...

def normInf(src: cv2.GMat) -> cv2.GScalar: ...

def normL1(src: cv2.GMat) -> cv2.GScalar: ...

def normL2(src: cv2.GMat) -> cv2.GScalar: ...

def normalize(src: cv2.GMat, alpha: float, beta: float, norm_type: int, ddepth: int = ...) -> cv2.GMat: ...

@typing.overload
def parseSSD(in_: cv2.GMat, inSz: cv2.GOpaqueT, confidenceThreshold: float = ..., filterLabel: int = ...) -> tuple[cv2.GArrayT, cv2.GArrayT]: ...
@typing.overload
def parseSSD(in_: cv2.GMat, inSz: cv2.GOpaqueT, confidenceThreshold: float, alignmentToSquare: bool, filterOutOfBounds: bool) -> cv2.GArrayT: ...

def parseYolo(in_: cv2.GMat, inSz: cv2.GOpaqueT, confidenceThreshold: float = ..., nmsThreshold: float = ..., anchors: typing.Sequence[float] = ...) -> tuple[cv2.GArrayT, cv2.GArrayT]: ...

def phase(x: cv2.GMat, y: cv2.GMat, angleInDegrees: bool = ...) -> cv2.GMat: ...

def polarToCart(magnitude: cv2.GMat, angle: cv2.GMat, angleInDegrees: bool = ...) -> tuple[cv2.GMat, cv2.GMat]: ...

def remap(src: cv2.GMat, map1: cv2.typing.MatLike, map2: cv2.typing.MatLike, interpolation: int, borderMode: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def resize(src: cv2.GMat, dsize: cv2.typing.Size, fx: float = ..., fy: float = ..., interpolation: int = ...) -> cv2.GMat: ...

def select(src1: cv2.GMat, src2: cv2.GMat, mask: cv2.GMat) -> cv2.GMat: ...

def sepFilter(src: cv2.GMat, ddepth: int, kernelX: cv2.typing.MatLike, kernelY: cv2.typing.MatLike, anchor: cv2.typing.Point, delta: cv2.typing.Scalar, borderType: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def split3(src: cv2.GMat) -> tuple[cv2.GMat, cv2.GMat, cv2.GMat]: ...

def split4(src: cv2.GMat) -> tuple[cv2.GMat, cv2.GMat, cv2.GMat, cv2.GMat]: ...

def sqrt(src: cv2.GMat) -> cv2.GMat: ...

def sub(src1: cv2.GMat, src2: cv2.GMat, ddepth: int = ...) -> cv2.GMat: ...

def subC(src: cv2.GMat, c: cv2.GScalar, ddepth: int = ...) -> cv2.GMat: ...

def subRC(c: cv2.GScalar, src: cv2.GMat, ddepth: int = ...) -> cv2.GMat: ...

def sum(src: cv2.GMat) -> cv2.GScalar: ...

@typing.overload
def threshold(src: cv2.GMat, thresh: cv2.GScalar, maxval: cv2.GScalar, type: int) -> cv2.GMat: ...
@typing.overload
def threshold(src: cv2.GMat, maxval: cv2.GScalar, type: int) -> tuple[cv2.GMat, cv2.GScalar]: ...

def transpose(src: cv2.GMat) -> cv2.GMat: ...

def warpAffine(src: cv2.GMat, M: cv2.typing.MatLike, dsize: cv2.typing.Size, flags: int = ..., borderMode: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...

def warpPerspective(src: cv2.GMat, M: cv2.typing.MatLike, dsize: cv2.typing.Size, flags: int = ..., borderMode: int = ..., borderValue: cv2.typing.Scalar = ...) -> cv2.GMat: ...





import sys
import time

from entrypoint2 import entrypoint

import pyscreenshot
from pyscreenshot.plugins.freedesktop_dbus import FreedesktopDBusWrapper
from pyscreenshot.plugins.gnome_dbus import GnomeDBusWrapper
from pyscreenshot.plugins.gnome_screenshot import GnomeScreenshotWrapper
from pyscreenshot.plugins.kwin_dbus import KwinDBusWrapper
from pyscreenshot.util import run_mod_as_subproc


def run(force_backend, n, childprocess, bbox=None):
    sys.stdout.write("%-20s\t" % force_backend)
    sys.stdout.flush()  # before any crash
    # if force_backend == "freedesktop_dbus":
    #     return
    if force_backend == "default":
        force_backend = None
    try:
        start = time.time()
        for _ in range(n):
            pyscreenshot.grab(
                backend=force_backend, childprocess=childprocess, bbox=bbox
            )
        end = time.time()
        dt = end - start
        s = "%-4.2g sec\t" % dt
        s += "(%5d ms per call)" % (1000.0 * dt / n)
        sys.stdout.write(s)
    finally:
        print("")


novirt = [
    GnomeDBusWrapper.name,
    KwinDBusWrapper.name,
    GnomeScreenshotWrapper.name,
    FreedesktopDBusWrapper.name,
]


def run_all(n, childprocess_param, virtual_only=True, bbox=None):
    debug = True
    print("")
    print("n=%s" % n)
    print("------------------------------------------------------")

    if bbox:
        x1, y1, x2, y2 = map(str, bbox)
        bbox = ":".join(map(str, (x1, y1, x2, y2)))
        bboxpar = ["--bbox", bbox]
    else:
        bboxpar = []
    if debug:
        debugpar = ["--debug"]
    else:
        debugpar = []
    for x in ["default"] + pyscreenshot.backends():
        if x == "freedesktop_dbus":
            continue
        backendpar = ["--backend", x]
        # skip non X backends
        if virtual_only and x in novirt:
            continue
        p = run_mod_as_subproc(
            "pyscreenshot.check.speedtest",
            ["--childprocess", childprocess_param, "--number", str(n)]
            + bboxpar
            + debugpar
            + backendpar,
        )
        print(p.stdout)


@entrypoint
def speedtest(virtual_display=False, backend="", childprocess="", bbox="", number=10):
    """Performance test of all back-ends.

    :param virtual_display: run with Xvfb
    :param bbox: bounding box coordinates x1:y1:x2:y2
    :param backend: back-end can be forced if set (example:default, scrot, wx,..),
                    otherwise all back-ends are tested
    :param childprocess: pyscreenshot parameter childprocess (0/1)
    :param number: number of screenshots for each backend (default:10)
    """
    childprocess_param = childprocess
    if childprocess == "":
        childprocess = True  # default
    elif childprocess == "0":
        childprocess = False
    elif childprocess == "1":
        childprocess = True
    else:
        raise ValueError("invalid childprocess value")

    if bbox:
        x1, y1, x2, y2 = map(int, bbox.split(":"))
        bbox = x1, y1, x2, y2
    else:
        bbox = None

    def f(virtual_only):
        if backend:
            try:
                run(backend, number, childprocess, bbox=bbox)
            except pyscreenshot.FailedBackendError:
                pass
        else:
            run_all(number, childprocess_param, virtual_only=virtual_only, bbox=bbox)

    if virtual_display:
        from pyvirtualdisplay import Display

        with Display(visible=0):
            f(virtual_only=True)
    else:
        f(virtual_only=False)



import threading
import time
import tomllib
from importlib import resources
from typing import Any, Callable, Dict

import pyautogui as gui
from pynput.keyboard import Key, KeyCode, Listener
from python_imagesearch.imagesearch import click_image, imagesearcharea

import ThreadWithTrace as thread_with_trace

_cfg = tomllib.loads(resources.read_text("game-bot-helper", "src/game-bot-helper/config.toml"))
exit_event = threading.Event()


def imagesearch_region_num_loop(image, time_sample, max_samples, x1, y1, x2, y2, precision=0.8):
    pos = imagesearcharea(image, x1, y1, x2, y2, precision)
    count = 0

    while pos[0] == -1:
        time.sleep(time_sample)
        pos = imagesearcharea(image, x1, y1, x2, y2, precision)
        count = count + 1
        if count > max_samples:
            break
    return pos


def click_image_modified(image: str, pos: [int, int], move_time: float = 1.0, wait_time_after: float = 1.0,
                         click_offset: int = 3):
    """Clicks on an image on screen with a bit of randomness.
    No imagesearch is performed, it is only needed to get the coordinates of the center of the image.

    :param image: path to the image file (see opencv imread for supported types)
    :param pos: the top left corner coordinates of the image as an array [x,y]
    :param move_time: time in seconds to move the mouse to the image
    :param wait_time_after: waiting time in seconds after clicking the image
    :param click_offset: offset in pixels from the center of the image to click

    """
    click_image(image, pos, "left", move_time, click_offset)
    time.sleep(wait_time_after)


class GameBotHelper:
    screen = None
    kill_switch_key = None
    current_threads = None

    # This will hold keys and their associated functions and arguments
    defined_runnable: Dict[Key | KeyCode, Dict[str, Any]] = {}

    def __init__(self):
        self.screen = _cfg['SCREEN']
        self.kill_switch_key = Key.end
        self.current_threads = []
        self.defined_runnable = {}

    def add_runnable(self, key: Key | KeyCode, runnable: Callable, args: list[Any] = None):
        """Adds a runnable to the bot.
        The runnable will be executed when the key is pressed.

        :param key: the key to press to execute the runnable. Can be a `Key` or `KeyCode`
        :param runnable: the function to execute without (), so do_stuff instead of do_stuff()
        :param args: the arguments to pass to the runnable
        """
        if args is None:
            args = []

        self.defined_runnable[key] = {"runnable": runnable, "args": args}

    def start(self):

        """Starts the bot.
        Will run until the kill switch key is pressed or your defined runnable finish.
        """
        listener = Listener(on_press=self.on_press, suppress=True)
        try:
            listener.start()
            print("Listener started")
            while not exit_event.is_set():
                # check if threads in list are done
                for thread in self.current_threads:
                    if not thread.is_alive():
                        self.current_threads.remove(thread)
                time.sleep(1)
            else:
                try:
                    gui.mouseUp()
                    for thread in self.current_threads:
                        thread.kill()
                        thread.join()
                        exit_event.clear()
                except AttributeError:
                    pass

        finally:
            listener.stop()
            listener.join()
            print("Listener stopped")
            return

    def on_press(self, key: Key | KeyCode):
        """Callback function for the key listener.
        Will execute the runnable associated with the key if it exists.
        """
        if key == self.kill_switch_key:
            print("Stopping all threads")
            exit_event.set()
            return

        if key in self.defined_runnable:
            print("Starting thread for " + str(key))
            print("Defined runnable: " + str(self.defined_runnable[key]))
            thread_to_start = thread_with_trace.ThreadWithTrace(target=self.defined_runnable[key]["runnable"],
                                                                args=self.defined_runnable[key]["args"])
            self.current_threads.append(thread_to_start)
            self.current_threads[-1].start()

    def change_kill_switch_key(self, key: Key | KeyCode):
        """Change the kill switch key. Default is Key.end"""
        self.kill_switch_key = key

    def change_screen(self, x1: int, y1: int, x2: int, y2: int, x_mid: int, y_mid: int):
        """Adjust screen coordinates to match the current game screen.
        This is only valid for the current instance of the GameBotHelper class!
        For a permanent change, edit the config.toml file.
        """
        self.screen['x1'] = x1
        self.screen['y1'] = y1
        self.screen['x2'] = x2
        self.screen['y2'] = y2
        self.screen['xMid'] = x_mid
        self.screen['yMid'] = y_mid

    def find_image(self, image: str, wait_time_between_tries: int, max_tries: int) -> [int, int]:
        """Searches for an image on screen continuously until it's found or max number of samples reached.

        :param image: path to the image file (see opencv imread for supported types)
        :param wait_time_between_tries: waiting time in seconds after failing to find the image
        :param max_tries: maximum number of samples before function times out.

        :returns: the top left corner coordinates of the element if found as an array [x,y] else [-1, -1]
        """
        return imagesearch_region_num_loop(image, wait_time_between_tries, max_tries, int(self.screen['x1']),
                                           int(self.screen['y1']),
                                           int(self.screen['x2']), int(self.screen['y2']), 0.8)

    def find_and_click_image(self, image: str, wait_time_between_tries: int = 1, max_tries: int = 5) -> bool:
        """Searches for an image on screen and clicks it.

        :param image: path to the image file (see opencv imread for supported types)
        :param wait_time_between_tries: waiting time in seconds after failing to find the image
        :param max_tries: maximum number of samples before function times out.

        :returns: True if the image was found and clicked, False otherwise
        """
        pos = self.find_image(image, wait_time_between_tries, max_tries)
        if pos[0] == -1:
            print("Image not found: " + image)
            return False
        click_image_modified(image, pos)
        return True



import time

from entrypoint2 import entrypoint

import pyscreenshot
from pyscreenshot import backends


@entrypoint
def show():
    im = []
    blist = []

    for x in backends():
        try:
            print("--> grabbing by " + x)
            im.append(pyscreenshot.grab(bbox=(500, 400, 800, 600), backend=x))
            blist.append(x)
        except Exception as e:
            print(e)
    print(im)
    print(blist)
    for x in im:
        x.show()
        time.sleep(0.5)




version = "1.26.1"
__version__ = version
full_version = version

git_revision = "411a55b9ec084c3315fe5f199351c31d0eb3c352"
release = 'dev' not in version and '+' not in version
short_version = version.split("+")[0]



#!/usr/bin/env python2.7

from __future__ import print_function

import fnmatch
import glob
import json
import os
import sys
import tempfile
import time
import types


from functools import wraps
from pprint import pprint

import click
import link_header
import requests
from requests import request


GITHUB_API = "https://api.github.com"

REQ_BUFFER_SIZE = 65536  # Chunk size when iterating a download body

_github_token_cli_arg = None


class _NoopProgressReporter(object):
    reportProgress = False

    def __init__(self, label='', length=0):
        self.label = label
        self.length = length

    def update(self, chunk_size):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass


progress_reporter_cls = _NoopProgressReporter
"""The progress reporter class to instantiate. This class
is expected to be a context manager with a constructor accepting `label`
and `length` keyword arguments, an `update` method accepting a `chunk_size`
argument and a class attribute `reportProgress` set to True (It can
conveniently be initialized using `sys.stdout.isatty()`)
"""


def _request(*args, **kwargs):
    with_auth = kwargs.pop("with_auth", True)
    token = _github_token_cli_arg
    if not token:
        token = os.environ.get("GITHUB_TOKEN", None)
    if token and with_auth:
        kwargs["auth"] = (token, 'x-oauth-basic')
    for _ in range(3):
        response = request(*args, **kwargs)
        is_travis = os.getenv("TRAVIS",  None) is not None
        if is_travis and 400 <= response.status_code < 500:
            print("Retrying in 1s (%s Client Error: %s for url: %s)" % (
                response.status_code, response.reason, response.url))
            time.sleep(1)
            continue
        break
    return response


def handle_http_error(func):
    @wraps(func)
    def with_error_handling(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.exceptions.HTTPError as e:
            print('Error sending {0} to {1}'.format(
                e.request.method, e.request.url))
            print('<', e.request.method, e.request.path_url)
            for k in sorted(e.request.headers.keys()):
                print('<', k, ':', e.request.headers[k])
            if e.request.body:
                print('<')
                print('<', repr(e.request.body[:35]),
                      '(total {0} bytes of data)'.format(len(e.request.body)))
            print('')
            print('>', e.response.status_code, e.response.reason)
            for k in sorted(e.response.headers.keys()):
                print('>', k.title(), ':', e.response.headers[k])
            if e.response.content:
                print('>')
                print('>', repr(e.response.content[:35]),
                      '(total {0} bytes of data)'.format(
                          len(e.response.content)))
            return 1
    return with_error_handling


def _check_for_credentials(func):
    @wraps(func)
    def with_check_for_credentials(*args, **kwargs):
        has_github_token_env_var = "GITHUB_TOKEN" in os.environ
        has_netrc = requests.utils.get_netrc_auth(GITHUB_API)
        if (not _github_token_cli_arg
                and not has_github_token_env_var and not has_netrc):
            raise EnvironmentError(
                "This command requires credentials provided by passing "
                "--github-token CLI argument, set using GITHUB_TOKEN "
                "env. variable or using netrc file. For more details, "
                "see https://github.com/j0057/github-release#configuring")
        return func(*args, **kwargs)
    return with_check_for_credentials


def _progress_bar(*args, **kwargs):
    bar = click.progressbar(*args, **kwargs)
    bar.bar_template = "  [%(bar)s]  %(info)s  %(label)s"
    bar.show_percent = True
    bar.show_pos = True

    def formatSize(length):
        if length == 0:
            return '%.2f' % length
        unit = ''
        # See https://en.wikipedia.org/wiki/Binary_prefix
        units = ['k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
        while True:
            if length <= 1024 or len(units) == 0:
                break
            unit = units.pop(0)
            length /= 1024.
        return '%.2f%s' % (length, unit)

    def formatPos(_self):
        pos = formatSize(_self.pos)
        if _self.length_known:
            pos += '/%s' % formatSize(_self.length)
        return pos

    bar.format_pos = types.MethodType(formatPos, bar)
    return bar


def _recursive_gh_get(href, items):
    """Recursively get list of GitHub objects.

    See https://developer.github.com/v3/guides/traversing-with-pagination/
    """
    response = _request('GET', href)
    response.raise_for_status()
    items.extend(response.json())
    if "link" not in response.headers:
        return
    links = link_header.parse(response.headers["link"])
    rels = {link.rel: link.href for link in links.links}
    if "next" in rels:
        _recursive_gh_get(rels["next"], items)


@click.group()
@click.option("--github-token", envvar='GITHUB_TOKEN', default=None,
              help="[default: GITHUB_TOKEN env. variable]")
@click.option("--progress/--no-progress", default=True,
              help="Display progress bar (default: yes).")
def main(github_token, progress):
    """A CLI to easily manage GitHub releases, assets and references."""
    global progress_reporter_cls
    progress_reporter_cls.reportProgress = sys.stdout.isatty() and progress
    if progress_reporter_cls.reportProgress:
        progress_reporter_cls = _progress_bar
    global _github_token_cli_arg
    _github_token_cli_arg = github_token


@main.group("release")
@click.argument('repo_name', metavar="REPOSITORY")
@click.pass_context
@handle_http_error
def gh_release(ctx, repo_name):
    """Manage releases (list, create, delete, ...) for
    REPOSITORY (e.g jcfr/sandbox)
    """
    ctx.obj = repo_name


# 1.6.0 (deprecated): Remove this bloc
class AssetGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        cmd_name = "delete" if cmd_name == "erase" else cmd_name
        return click.Group.get_command(self, ctx, cmd_name)


@main.group("asset", cls=AssetGroup)
@click.argument('repo_name', metavar="REPOSITORY")
@click.pass_context
@handle_http_error
def gh_asset(ctx, repo_name):
    """Manage release assets (upload, download, ...) for
    REPOSITORY (e.g jcfr/sandbox)
    """
    ctx.obj = repo_name


@main.group("ref")
@click.argument('repo_name', metavar="REPOSITORY")
@click.pass_context
@handle_http_error
def gh_ref(ctx, repo_name):
    """Manage references (list, create, delete, ...) for
    REPOSITORY (e.g jcfr/sandbox)
    """
    ctx.obj = repo_name


#
# Releases
#

def print_release_info(release, title=None, indent=""):
    if title is None:
        title = "release '{0}' info".format(release["tag_name"])
    print(indent + title)
    indent = "  " + indent
    print(indent + 'Tag name      : {tag_name}'.format(**release))
    if release['name']:
        print(indent + 'Name          : {name}'.format(**release))
    print(indent + 'ID            : {id}'.format(**release))
    print(indent + 'Created       : {created_at}'.format(**release))
    print(indent + 'URL           : {html_url}'.format(**release))
    print(indent + 'Author        : {login}'.format(**release['author']))
    print(indent + 'Is published  : {0}'.format(not release['draft']))
    print(indent + 'Is prerelease : {0}'.format(release['prerelease']))
    if release['body']:
        print(indent + 'Release notes :')
        print(indent + release['body'])
    print('')
    for (i, asset) in enumerate(release['assets']):
        print_asset_info(i, asset, indent=indent)


def get_releases(repo_name, verbose=False):

    releases = []
    _recursive_gh_get(
        GITHUB_API + '/repos/{0}/releases'.format(repo_name), releases)

    if verbose:
        list(map(print_release_info,
                 sorted(releases, key=lambda r: r['tag_name'])))
    return releases


def get_release(repo_name, tag_name):
    releases = get_releases(repo_name)
    try:
        release = next(r for r in releases if r['tag_name'] == tag_name)
        return release
    except StopIteration:
        return None


def get_release_info(repo_name, tag_name):
    release = get_release(repo_name, tag_name)
    if release is not None:
        return release
    else:
        raise Exception('Release with tag_name {0} not found'.format(tag_name))


def _update_release_sha(repo_name, tag_name, new_release_sha, dry_run):
    """Update the commit associated with a given release tag.

    Since updating a tag commit is not directly possible, this function
    does the following steps:
    * set the release tag to ``<tag_name>-tmp`` and associate it
      with ``new_release_sha``.
    * delete tag ``refs/tags/<tag_name>``.
    * update the release tag to ``<tag_name>`` and associate it
      with ``new_release_sha``.
    """
    if new_release_sha is None:
        return
    refs = get_refs(repo_name, tags=True, pattern="refs/tags/%s" % tag_name)
    if not refs:
        return
    assert len(refs) == 1

    # If sha associated with "<tag_name>" is up-to-date, we are done.
    previous_release_sha = refs[0]["object"]["sha"]
    if previous_release_sha == new_release_sha:
        return

    tmp_tag_name = tag_name + "-tmp"

    # If any, remove leftover temporary tag "<tag_name>-tmp"
    refs = get_refs(repo_name, tags=True, pattern="refs/tags/%s" % tmp_tag_name)
    if refs:
        assert len(refs) == 1
        time.sleep(0.1)
        gh_ref_delete(repo_name,
                      "refs/tags/%s" % tmp_tag_name, dry_run=dry_run)

    # Update "<tag_name>" release by associating it with the "<tag_name>-tmp"
    # and "<new_release_sha>". It will create the temporary tag.
    time.sleep(0.1)
    patch_release(repo_name, tag_name,
                  tag_name=tmp_tag_name,
                  target_commitish=new_release_sha,
                  dry_run=dry_run)

    # Now "<tag_name>-tmp" references "<new_release_sha>", remove "<tag_name>"
    time.sleep(0.1)
    gh_ref_delete(repo_name, "refs/tags/%s" % tag_name, dry_run=dry_run)

    # Finally, update "<tag_name>-tmp" release by associating it with the
    # "<tag_name>" and "<new_release_sha>".
    time.sleep(0.1)
    patch_release(repo_name, tmp_tag_name,
                  tag_name=tag_name,
                  target_commitish=new_release_sha,
                  dry_run=dry_run)

    # ... and remove "<tag_name>-tmp"
    time.sleep(0.1)
    gh_ref_delete(repo_name,
                  "refs/tags/%s" % tmp_tag_name, dry_run=dry_run)


def patch_release(repo_name, current_tag_name, **values):
    dry_run = values.get("dry_run", False)
    verbose = values.get("verbose", False)
    release = get_release_info(repo_name, current_tag_name)
    new_tag_name = values.get("tag_name", release["tag_name"])

    _update_release_sha(
        repo_name,
        new_tag_name,
        values.get("target_commitish", None),
        dry_run
    )

    data = {
        "tag_name": release["tag_name"],
        "target_commitish": release["target_commitish"],
        "name": release["name"],
        "body": release["body"],
        "draft": release["draft"],
        "prerelease": release["prerelease"]
    }

    updated = []
    for key in data:
        if key in values and data[key] != values[key]:
            updated.append("%s: '%s' -> '%s'" % (key, data[key], values[key]))
    if updated:
        print("updating '%s' release: \n  %s" % (
            current_tag_name, "\n  ".join(updated)))
        print("")

    data.update(values)

    if not dry_run:
        url = GITHUB_API + '/repos/{0}/releases/{1}'.format(
            repo_name, release['id'])
        response = _request(
            'PATCH', url,
            data=json.dumps(data),
            headers={'Content-Type': 'application/json'})
        response.raise_for_status()

    # In case a new tag name was provided, remove the old one.
    if current_tag_name != data["tag_name"]:
        gh_ref_delete(
            repo_name, "refs/tags/%s" % current_tag_name,
            tags=True, verbose=verbose, dry_run=dry_run)


def get_assets(repo_name, tag_name, verbose=False):
    release = get_release(repo_name, tag_name)
    if not release:
        raise Exception('Release with tag_name {0} not found'.format(tag_name))

    assets = []
    _recursive_gh_get(GITHUB_API + '/repos/{0}/releases/{1}/assets'.format(
        repo_name, release["id"]), assets)

    if verbose:
        for i, asset in enumerate(sorted(assets, key=lambda r: r['name'])):
            print_asset_info(i, asset)

    return assets


def get_asset_info(repo_name, tag_name, filename):
    assets = get_assets(repo_name, tag_name)
    try:
        asset = next(a for a in assets if a['name'] == filename)
        return asset
    except StopIteration:
        raise Exception('Asset with filename {0} not found in '
                        'release with tag_name {1}'.format(filename, tag_name))


@gh_release.command("list")
@click.pass_obj
def _cli_release_list(repo_name):
    """List releases"""
    return get_releases(repo_name, verbose=True)


@gh_release.command("info")
@click.argument("tag_name")
@click.pass_obj
def _cli_release_info(repo_name, tag_name):
    """Get release description"""
    release = get_release_info(repo_name, tag_name)
    print_release_info(release)


@gh_release.command("create")
@click.argument("tag_name")
@click.argument("asset_pattern", nargs=-1)
@click.option("--name")
@click.option("--body", default=None)
@click.option("--publish", is_flag=True, default=False)
@click.option("--prerelease", is_flag=True, default=False)
@click.option("--dry-run", is_flag=True, default=False)
@click.option("--target-commitish")
@click.pass_obj
def cli_release_create(*args, **kwargs):
    """Create a release"""
    gh_release_create(*args, **kwargs)


@_check_for_credentials
def gh_release_create(repo_name, tag_name, asset_pattern=None, name=None, body=None,
                      publish=False, prerelease=False,
                      target_commitish=None, dry_run=False):
    if get_release(repo_name, tag_name) is not None:
        print('release %s: already exists\n' % tag_name)
        return
    data = {
        'tag_name': tag_name,
        'draft': not publish and not prerelease,
        'prerelease': prerelease
    }
    if name is not None:
        data["name"] = name
    if body is not None:
        data["body"] = body
    if target_commitish is not None:
        data["target_commitish"] = target_commitish
    if not dry_run:
        response = _request(
              'POST', GITHUB_API + '/repos/{0}/releases'.format(repo_name),
              data=json.dumps(data),
              headers={'Content-Type': 'application/json'})
        response.raise_for_status()
        print_release_info(response.json(),
                           title="created '%s' release" % tag_name)
    else:
        print("created '%s' release (dry_run)" % tag_name)
    if asset_pattern is not None:
        gh_asset_upload(repo_name, tag_name, asset_pattern, dry_run=dry_run)


@gh_release.command("edit")
@click.argument("current_tag_name")
@click.option("--tag-name", default=None)
@click.option("--target-commitish", default=None)
@click.option("--name", default=None)
@click.option("--body", default=None)
@click.option("--draft/--publish", is_flag=True, default=None)
@click.option("--prerelease/--release", is_flag=True, default=None)
@click.option("--dry-run", is_flag=True, default=False)
@click.option("--verbose", is_flag=True, default=False)
@click.pass_obj
def _cli_release_edit(*args, **kwargs):
    """Edit a release"""
    gh_release_edit(*args, **kwargs)


@_check_for_credentials
def gh_release_edit(repo_name, current_tag_name,
                    tag_name=None, target_commitish=None, name=None,
                    body=None,
                    draft=None, prerelease=None, dry_run=False, verbose=False):
    attributes = {}
    for key in [
        "tag_name", "target_commitish", "name", "body", "draft",
        "prerelease", "dry_run", "verbose"
    ]:
        if locals().get(key, None) is not None:
            attributes[key] = locals()[key]
    patch_release(repo_name, current_tag_name, **attributes)


@gh_release.command("delete")
@click.argument("pattern")
@click.option("--keep-pattern")
@click.option("--dry-run", is_flag=True, default=False)
@click.option("--verbose", is_flag=True, default=False)
@click.pass_obj
def _cli_release_delete(*args, **kwargs):
    """Delete selected release"""
    gh_release_delete(*args, **kwargs)


@_check_for_credentials
def gh_release_delete(repo_name, pattern, keep_pattern=None,
                      dry_run=False, verbose=False):
    releases = get_releases(repo_name)
    candidates = []
    # Get list of candidate releases
    for release in releases:
        if not fnmatch.fnmatch(release['tag_name'], pattern):
            if verbose:
                print('skipping release {0}: do not match {1}'.format(
                    release['tag_name'], pattern))
            continue
        if keep_pattern is not None:
            if fnmatch.fnmatch(release['tag_name'], keep_pattern):
                continue
        candidates.append(release['tag_name'])
    for tag_name in candidates:
        release = get_release(repo_name, tag_name)
        print('deleting release {0}'.format(release['tag_name']))
        if dry_run:
            continue
        url = (GITHUB_API
               + '/repos/{0}/releases/{1}'.format(repo_name, release['id']))
        response = _request('DELETE', url)
        response.raise_for_status()
    return len(candidates) > 0


@gh_release.command("publish")
@click.argument("tag_name")
@click.option("--prerelease", is_flag=True, default=False)
@click.pass_obj
def _cli_release_publish(*args, **kwargs):
    """Publish a release setting draft to 'False'"""
    gh_release_publish(*args, **kwargs)


@_check_for_credentials
def gh_release_publish(repo_name, tag_name, prerelease=False):
    patch_release(repo_name, tag_name, draft=False, prerelease=prerelease)


@gh_release.command("unpublish")
@click.argument("tag_name")
@click.option("--prerelease", is_flag=True, default=False)
@click.pass_obj
def _cli_release_unpublish(*args, **kwargs):
    """Unpublish a release setting draft to 'True'"""
    gh_release_unpublish(*args, **kwargs)


@_check_for_credentials
def gh_release_unpublish(repo_name, tag_name, prerelease=False):
    draft = not prerelease
    patch_release(repo_name, tag_name, draft=draft, prerelease=prerelease)


@gh_release.command("release-notes")
@click.argument("tag_name")
@click.pass_obj
def _cli_release_notes(*args, **kwargs):
    """Set release notes"""
    gh_release_notes(*args, **kwargs)


@_check_for_credentials
def gh_release_notes(repo_name, tag_name):
    release = get_release_info(repo_name, tag_name)
    (_, filename) = tempfile.mkstemp(suffix='.md')
    try:
        if release['body']:
            with open(filename, 'w+b') as f:
                f.write(release['body'])
        ret = os.system('{0} {1}'.format(os.environ['EDITOR'], filename))
        if ret:
            raise Exception(
                '{0} returned exit code {1}'.format(os.environ['EDITOR'], ret))
        with open(filename, 'rb') as f:
            body = f.read()
        if release['body'] == body:
            return
        patch_release(repo_name, tag_name, body=body)
    finally:
        os.remove(filename)


@gh_release.command("debug")
@click.argument("tag_name")
@click.pass_obj
def _cli_release_debug(repo_name, tag_name):
    """Print release detailed information"""
    release = get_release_info(repo_name, tag_name)
    pprint(release)


#
# Assets
#

def print_asset_info(i, asset, indent=""):
    print(indent + "Asset #{i}".format(i=i))
    indent = "  " + indent
    print(indent + "name      : {name}".format(i=i, **asset))
    print(indent + "state     : {state}".format(i=i, **asset))
    print(indent + "uploader  : {login}".format(i=i, **asset['uploader']))
    print(indent + "size      : {size}".format(i=i, **asset))
    print(indent + "URL       : {browser_download_url}".format(i=i, **asset))
    print(indent + "Downloads : {download_count}".format(i=i, **asset))
    print("")


@gh_asset.command("upload")
@click.argument("tag_name")
@click.argument("pattern", nargs=-1)
@click.pass_obj
def _cli_asset_upload(*args, **kwargs):
    """Upload release assets"""
    gh_asset_upload(*args, **kwargs)


class _ProgressFileReader(object):
    """Wrapper used to capture File IO read progress."""
    def __init__(self, stream, reporter):
        self._stream = stream
        self._reporter = reporter

    def read(self, _size):
        _chunk = self._stream.read(_size)
        self._reporter.update(len(_chunk))
        return _chunk

    def __getattr__(self, attr):
        return getattr(self._stream, attr)


def _upload_release_file(
        repo_name, tag_name, upload_url, filename,
        verbose=False, dry_run=False, retry=True):
    already_uploaded = False
    uploaded = False
    basename = os.path.basename(filename)
    # Sanity checks
    assets = get_assets(repo_name, tag_name)
    download_url = None
    for asset in assets:
        if asset["name"] == basename:
            if asset["state"] == "uploaded":
                download_url = asset["browser_download_url"]
                break
            # Remove asset that failed to upload
            # See https://developer.github.com/v3/repos/releases/#response-for-upstream-failure  # noqa: E501
            if asset["state"] == "new":
                print("  deleting %s (invalid asset "
                      "with state set to 'new')" % asset['name'])
                url = (
                    GITHUB_API
                    + '/repos/{0}/releases/assets/{1}'.format(
                        repo_name, asset['id'])
                )
                response = _request('DELETE', url)
                response.raise_for_status()

    print("  uploading %s" % filename)

    # Skip if an asset with same name has already been uploaded
    # Trying to upload would give a HTTP error 422
    if download_url:
        already_uploaded = True
        print("  skipping (asset with same name already exists)")
        print("  download_url: %s" % download_url)
        print("")
        return already_uploaded, uploaded, {}
    if dry_run:
        uploaded = True
        print("  download_url: Unknown (dry_run)")
        print("")
        return already_uploaded, uploaded, {}

    url = '{0}?name={1}'.format(upload_url, basename)
    if verbose and not progress_reporter_cls.reportProgress:
        print("  upload_url: %s" % url)
    file_size = os.path.getsize(filename)

    # Attempt upload
    with open(filename, 'rb') as f:
        with progress_reporter_cls(
                label=basename, length=file_size) as reporter:
            response = _request(
                'POST', url,
                headers={'Content-Type': 'application/octet-stream'},
                data=_ProgressFileReader(f, reporter))
            data = response.json()

    if response.status_code == 502 and retry:
        print("  retrying (upload failed with status_code=502)")
        already_uploaded, uploaded, data = _upload_release_file(
            repo_name, tag_name, upload_url, filename,
            verbose=verbose, retry=False)
    else:
        response.raise_for_status()
    asset = data
    print("  download_url: %s" % asset["browser_download_url"])
    print("")
    uploaded = True
    return already_uploaded, uploaded, response.json()


@_check_for_credentials
def gh_asset_upload(repo_name, tag_name, pattern, dry_run=False, verbose=False):
    if not dry_run:
        upload_url = get_release_info(repo_name, tag_name)["upload_url"]
        if "{" in upload_url:
            upload_url = upload_url[:upload_url.index("{")]
    else:
        upload_url = "unknown"

    # Raise exception if no token is specified AND netrc file is found
    # BUT only api.github.com is specified. See #17
    has_github_token = "GITHUB_TOKEN" in os.environ
    has_netrc = requests.utils.get_netrc_auth(GITHUB_API)
    if not has_github_token and has_netrc:
        if requests.utils.get_netrc_auth(upload_url) is None:
            raise EnvironmentError(
                "Found netrc file but upload URL is missing. "
                "For more details, "
                "see https://github.com/j0057/github-release#configuring")

    if type(pattern) in [list, tuple]:
        filenames = []
        for package in pattern:
            filenames.extend(glob.glob(package))
        filenames = set(filenames)
    elif pattern:
        filenames = glob.glob(pattern)
    else:
        filenames = []

    if len(filenames) > 0:
        print("uploading '%s' release asset(s) "
              "(found %s):" % (tag_name, len(filenames)))

    uploaded = False
    already_uploaded = False

    for filename in filenames:
        already_uploaded, uploaded, _ = _upload_release_file(
            repo_name, tag_name, upload_url, filename, verbose, dry_run)

    if not uploaded and not already_uploaded:
        print("skipping upload of '%s' release assets ("
              "no files match pattern(s): %s)" % (tag_name, pattern))
        print("")


@gh_asset.command("delete")
@click.argument("tag_name")
@click.argument("pattern")
@click.option("--keep-pattern", default=None)
@click.option("--dry-run", is_flag=True, default=False)
@click.option("--verbose", is_flag=True, default=False)
@click.pass_obj
def _cli_asset_delete(*args, **kwargs):
    """Delete selected release assets"""
    gh_asset_delete(*args, **kwargs)


@_check_for_credentials
def gh_asset_delete(repo_name, tag_name, pattern,
                    keep_pattern=None, dry_run=False, verbose=False):
    # Get assets
    assets = get_assets(repo_name, tag_name)
    # List of assets
    excluded_assets = {}
    matched_assets = []
    matched_assets_to_keep = {}
    for asset in assets:
        if not fnmatch.fnmatch(asset['name'], pattern):
            skip_reason = "do NOT match pattern '%s'" % pattern
            excluded_assets[asset['name']] = skip_reason
            continue
        matched_assets.append(asset)
        if keep_pattern is not None:
            if fnmatch.fnmatch(asset['name'], keep_pattern):
                skip_reason = "match keep_pattern '%s'" % keep_pattern
                matched_assets_to_keep[asset['name']] = skip_reason
                continue
    # Summary
    summary = "matched: %s, matched-but-keep: %s, not-matched: %s" % (
        len(matched_assets),
        len(matched_assets_to_keep),
        len(excluded_assets)
    )
    print("deleting '%s' release asset(s) (%s):" % (tag_name, summary))
    # Perform deletion
    for asset in matched_assets:
        if asset['name'] in matched_assets_to_keep:
            if verbose:
                skip_reason = matched_assets_to_keep[asset['name']]
                print("  skipping %s (%s)" % (asset['name'], skip_reason))
            continue
        print("  deleting %s" % asset['name'])
        if dry_run:
            continue
        url = (
            GITHUB_API
            + '/repos/{0}/releases/assets/{1}'.format(repo_name, asset['id'])
        )
        response = _request('DELETE', url)
        response.raise_for_status()
    if len(matched_assets) == 0:
        print("  nothing to delete")
    print("")
    if verbose:
        indent = "  "
        print(indent + "assets NOT matching selection pattern [%s]:" % pattern)
        for asset_name in excluded_assets:
            print(indent + "  " + asset_name)
        print("")


@gh_asset.command("download")
@click.argument("tag_name")
@click.argument("pattern", required=False)
@click.pass_obj
def _cli_asset_download(*args, **kwargs):
    """Download release assets"""
    gh_asset_download(*args, **kwargs)


def _download_file(repo_name, asset):
    response = _request(
        method='GET',
        url=GITHUB_API + '/repos/{0}/releases/assets/{1}'.format(
            repo_name, asset['id']),
        allow_redirects=False,
        headers={'Accept': 'application/octet-stream'},
        stream=True)
    while response.status_code == 302:
        response = _request(
            'GET', response.headers['Location'], allow_redirects=False,
            stream=True,
            with_auth=False
        )
    with open(asset['name'], 'w+b') as f:
        with progress_reporter_cls(
                label=asset['name'], length=asset['size']) as reporter:
            for chunk in response.iter_content(chunk_size=REQ_BUFFER_SIZE):
                reporter.update(len(chunk))
                f.write(chunk)


def gh_asset_download(repo_name, tag_name=None, pattern=None):
    releases = get_releases(repo_name)
    downloaded = 0
    for release in releases:
        if tag_name and not fnmatch.fnmatch(release['tag_name'], tag_name):
            continue
        for asset in release['assets']:
            if pattern and not fnmatch.fnmatch(asset['name'], pattern):
                continue
            if os.path.exists(asset['name']):
                absolute_path = os.path.abspath(asset['name'])
                print('release {0}: '
                      'skipping {1}: '
                      'found {2}'.format(
                        release['tag_name'], asset['name'], absolute_path))
                continue
            print('release {0}: '
                  'downloading {1}'.format(release['tag_name'], asset['name']))
            _download_file(repo_name, asset)
            downloaded += 1
    return downloaded


@gh_asset.command("list")
@click.argument("tag_name")
@click.pass_obj
def _cli_asset_list(repo_name, tag_name):
    """List release assets"""
    return get_assets(repo_name, tag_name, verbose=True)


#
# References
#

def print_object_info(ref_object, indent=""):
    print(indent + 'Object')
    print(indent + '  type : {type}'.format(**ref_object))
    print(indent + '  sha  : {sha}'.format(**ref_object))


def print_ref_info(ref, indent=""):
    print(indent + "Reference '{ref}'".format(**ref))
    print_object_info(ref['object'], indent="  " + indent)
    print("")


def get_refs(repo_name, tags=None, pattern=None):

    refs = []
    _recursive_gh_get(
        GITHUB_API + '/repos/{0}/git/refs'.format(repo_name), refs)

    # If "tags" is True, keep only "refs/tags/*"
    data = refs
    if tags:
        tag_names = []
        data = []
        for ref in refs:
            if ref['ref'].startswith("refs/tags"):
                data.append(ref)
                tag_names.append(ref["ref"])

        try:
            tags = []
            _recursive_gh_get(
                GITHUB_API + '/repos/{0}/git/refs/tags'.format(repo_name), tags)
            for ref in tags:
                if ref["ref"] not in tag_names:
                    data.append(ref)
        except requests.exceptions.HTTPError as exc_info:
            response = exc_info.response
            if response.status_code != 404:
                raise

    # If "pattern" is not None, select only matching references
    filtered_data = data
    if pattern is not None:
        filtered_data = []
        for ref in data:
            if fnmatch.fnmatch(ref['ref'], pattern):
                filtered_data.append(ref)

    return filtered_data


@gh_ref.command("list")
@click.option("--tags", is_flag=True, default=False)
@click.option("--pattern", default=None)
@click.option("--verbose", is_flag=True, default=False)
@click.pass_obj
def _cli_ref_list(*args, **kwargs):
    """List all references"""
    gh_ref_list(*args, **kwargs)


def gh_ref_list(repo_name, tags=None,  pattern=None, verbose=False):
    refs = get_refs(repo_name, tags=tags, pattern=pattern)
    sorted_refs = sorted(refs, key=lambda r: r['ref'])
    if verbose:
        list(map(print_ref_info, sorted_refs))
    else:
        list(map(lambda ref: print(ref['ref']), sorted_refs))
    return sorted_refs


@gh_ref.command("create")
@click.argument("reference")
@click.argument("sha")
@click.pass_obj
def _cli_ref_create(*args, **kwargs):
    """Create reference (e.g heads/foo, tags/foo)"""
    gh_ref_create(*args, **kwargs)


@_check_for_credentials
def gh_ref_create(repo_name, reference, sha):
    data = {
        'ref': "refs/%s" % reference,
        'sha': sha
    }
    response = _request(
          'POST', GITHUB_API + '/repos/{0}/git/refs'.format(repo_name),
          data=json.dumps(data),
          headers={'Content-Type': 'application/json'})
    response.raise_for_status()
    print_ref_info(response.json())


@gh_ref.command("delete")
@click.argument("pattern")
@click.option("--keep-pattern", default=None)
@click.option("--tags", is_flag=True, default=False)
@click.option("--dry-run", is_flag=True, default=False)
@click.option("--verbose", is_flag=True, default=False)
@click.pass_obj
def _cli_ref_delete(*args, **kwargs):
    """Delete selected references"""
    gh_ref_delete(*args, **kwargs)


@_check_for_credentials
def gh_ref_delete(repo_name, pattern, keep_pattern=None, tags=False,
                  dry_run=False, verbose=False):
    removed_refs = []
    refs = get_refs(repo_name, tags=tags)
    for ref in refs:
        if not fnmatch.fnmatch(ref['ref'], pattern):
            if verbose:
                print('skipping reference {0}: '
                      'do not match {1}'.format(ref['ref'], pattern))
            continue
        if keep_pattern is not None:
            if fnmatch.fnmatch(ref['ref'], keep_pattern):
                continue
        print('deleting reference {0}'.format(ref['ref']))
        removed_refs.append(ref['ref'])
        if dry_run:
            continue
        response = _request(
            'DELETE',
            GITHUB_API + '/repos/{0}/git/{1}'.format(repo_name, ref['ref']))
        response.raise_for_status()
    return len(removed_refs) > 0


#
# Commits
#

def gh_commit_get(repo_name, sha):
    try:
        response = _request(
            'GET',
            GITHUB_API + '/repos/{0}/git/commits/{1}'.format(repo_name, sha))
        response.raise_for_status()
        return response.json()
    except requests.exceptions.HTTPError as exc_info:
        response = exc_info.response
        if response.status_code == 404:
            return None
        else:
            raise


#
# Script entry point
#

if __name__ == '__main__':
    main()



"""
Backend for grim (https://github.com/emersion/grim), a Wayland screen tool for
environments other than Gnome and KDE, such as Sway.
"""
import logging

from easyprocess import EasyProcess

from pyscreenshot.plugins.backend import UNKNOWN_VERSION, CBackend
from pyscreenshot.tempexport import read_prog_img

log = logging.getLogger(__name__)


PROGRAM = "grim"


class GrimWrapper(CBackend):
    name = "grim"
    is_subprocess = True

    def _bbox_to_grim_region(self, bbox):
        """
        Translate pyscreenshot's bbox tuple convention of (x1, y1, x2, y2) to
        grim's bbox convention, which is a string of the following format:
        <x>,<y> <width>x<height>
        """
        x1, y1, x2, y2 = bbox
        width = x2 - x1
        height = y2 - y1
        return "{},{} {}x{}".format(x1, y1, width, height)

    def grab(self, bbox=None):
        if bbox:
            # using grim's built-in cropping feature
            region = self._bbox_to_grim_region(bbox)
            return read_prog_img([PROGRAM, "-g", region])
        return read_prog_img([PROGRAM])

    def backend_version(self):
        # grim doesn't have a version flag for some reason
        p = EasyProcess([PROGRAM, "-help"])
        p.enable_stdout_log = False
        p.enable_stderr_log = False
        p.call()
        if p.return_code == 0:
            return UNKNOWN_VERSION



import logging
import traceback

from pyscreenshot.childproc import childprocess_grab
from pyscreenshot.err import FailedBackendError
from pyscreenshot.plugins.freedesktop_dbus import FreedesktopDBusWrapper
from pyscreenshot.plugins.gdk3pixbuf import Gdk3PixbufWrapper
from pyscreenshot.plugins.gnome_dbus import GnomeDBusWrapper
from pyscreenshot.plugins.gnome_screenshot import GnomeScreenshotWrapper
from pyscreenshot.plugins.grim import GrimWrapper
from pyscreenshot.plugins.imagemagick import ImagemagickWrapper

# from pyscreenshot.plugins.kwin_dbus import KwinDBusWrapper
from pyscreenshot.plugins.mac_quartz import MacQuartzWrapper
from pyscreenshot.plugins.mac_screencapture import ScreencaptureWrapper
from pyscreenshot.plugins.maim import MaimWrapper
from pyscreenshot.plugins.msswrap import MssWrapper
from pyscreenshot.plugins.pilwrap import PilWrapper
from pyscreenshot.plugins.pyside2_grabwindow import PySide2GrabWindow

# from pyscreenshot.plugins.pyside_grabwindow import PySideGrabWindow
# from pyscreenshot.plugins.qt4grabwindow import Qt4GrabWindow
from pyscreenshot.plugins.qt5grabwindow import Qt5GrabWindow
from pyscreenshot.plugins.scrot import ScrotWrapper
from pyscreenshot.plugins.wxscreen import WxScreen

# from pyscreenshot.plugins.ksnip import KsnipWrapper
from pyscreenshot.util import (
    platform_is_linux,
    platform_is_osx,
    platform_is_win,
    use_x_display,
)

log = logging.getLogger(__name__)


backend_dict = {
    PilWrapper.name: PilWrapper,
    MssWrapper.name: MssWrapper,
    ScrotWrapper.name: ScrotWrapper,
    GrimWrapper.name: GrimWrapper,
    MaimWrapper.name: MaimWrapper,
    ImagemagickWrapper.name: ImagemagickWrapper,
    Qt5GrabWindow.name: Qt5GrabWindow,
    # Qt4GrabWindow.name: Qt4GrabWindow,
    PySide2GrabWindow.name: PySide2GrabWindow,
    # PySideGrabWindow.name: PySideGrabWindow,
    WxScreen.name: WxScreen,
    Gdk3PixbufWrapper.name: Gdk3PixbufWrapper,
    ScreencaptureWrapper.name: ScreencaptureWrapper,
    MacQuartzWrapper.name: MacQuartzWrapper,
    FreedesktopDBusWrapper.name: FreedesktopDBusWrapper,
    GnomeDBusWrapper.name: GnomeDBusWrapper,
    GnomeScreenshotWrapper.name: GnomeScreenshotWrapper,
    # KwinDBusWrapper.name: KwinDBusWrapper,
    # XwdWrapper.name: XwdWrapper,
    # KsnipWrapper.name: KsnipWrapper,
}


def qt():
    yield Qt5GrabWindow
    # yield Qt4GrabWindow
    yield PySide2GrabWindow
    # yield PySideGrabWindow


def backends(childprocess):
    # the order is based on performance
    if platform_is_linux():
        if use_x_display():
            if childprocess:
                yield ScrotWrapper
                yield PilWrapper
                yield MssWrapper
            else:
                yield PilWrapper
                yield MssWrapper
                yield ScrotWrapper
            yield MaimWrapper
            yield ImagemagickWrapper
            yield Gdk3PixbufWrapper
            yield WxScreen
            for x in qt():
                yield x

        yield GnomeDBusWrapper

        # on screen notification
        # "The process is not authorized to take a screenshot"
        # yield KwinDBusWrapper

        # flash effect
        yield GnomeScreenshotWrapper

        yield GrimWrapper
        # yield KsnipWrapper

        # confirmation dialog
        yield FreedesktopDBusWrapper

    elif platform_is_osx():
        yield PilWrapper
        yield MssWrapper

        # alternatives for older pillow versions
        yield ScreencaptureWrapper
        yield MacQuartzWrapper

        # qt has some color difference

        # does not work: Gdk3, wx, Imagemagick

    elif platform_is_win():
        yield PilWrapper
        yield MssWrapper
    else:
        yield PilWrapper
        yield MssWrapper
        for x in backend_dict.values():
            yield x


def select_childprocess(childprocess, backend_class):
    if backend_class.is_subprocess:
        # backend is always a subprocess -> nothing to do
        return False

    return childprocess


def auto(bbox, childprocess):
    im = None
    for backend_class in backends(childprocess):
        backend_name = backend_class.name
        try:
            if select_childprocess(childprocess, backend_class):
                log.debug('running "%s" in child process', backend_name)
                im = childprocess_grab(backend_name, bbox)
            else:
                obj = backend_class()

                im = obj.grab(bbox)
            break
        except Exception:
            msg = traceback.format_exc()
            log.debug(msg)
    if not im:
        msg = "All backends failed!"
        raise FailedBackendError(msg)

    return im


def force(backend_name, bbox, childprocess):
    backend_class = backend_dict[backend_name]
    if select_childprocess(childprocess, backend_class):
        log.debug('running "%s" in child process', backend_name)
        return childprocess_grab(backend_name, bbox)
    else:
        obj = backend_class()
        im = obj.grab(bbox)
        return im


def backend_grab(backend, bbox, childprocess):
    if backend:
        return force(backend, bbox, childprocess)
    else:
        return auto(bbox, childprocess)


def backend_version2(backend_name):
    backend_class = backend_dict[backend_name]
    obj = backend_class()
    v = obj.backend_version()
    return v
